# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Callable, Optional, Union, List, Tuple, get_origin, get_args
import numpy.typing as npt
import numpy as np
import pandas as pd
import math
import inspect
from functools import partial
import warnings

from boario import logger
from boario.utils.recovery_functions import (
    concave_recovery,
    convexe_recovery,
    linear_recovery,
    convexe_recovery_scaled,
)

__all__ = [
    "Event",
    "EventKapitalDestroyed",
    "EventArbitraryProd",
    "EventKapitalRecover",
    "EventKapitalRebuild",
    "Impact",
    "IndustriesList",
    "SectorsList",
    "RegionsList",
]

VectorImpact = Union[list, dict, np.ndarray, pd.DataFrame, pd.Series]
ScalarImpact = Union[int, float, np.integer]
Impact = Union[VectorImpact, ScalarImpact]
IndustriesList = Union[List[Tuple[str, str]], pd.MultiIndex]
SectorsList = Union[List[str], pd.Index, str]
RegionsList = Union[List[str], pd.Index, str]
FinalCatList = Union[List[str], pd.Index, str]

REBUILDING_FINALDEMAND_CAT_REGEX = (
    r"(?i)(?=.*household)(?=.*final)(?!.*NPISH|profit).*|HFCE"
)

LOW_DEMAND_THRESH = 10


class Event(ABC):
    # Class Attributes

    possible_sectors: pd.Index = pd.Index([])
    r"""List of sectors present in the MRIO used by the model"""

    possible_regions: pd.Index = pd.Index([])
    r"""List of regions present in the MRIO used by the model"""

    possible_final_demand_cat: pd.Index = pd.Index([])
    r"""List of final demand categories present in the MRIO used by the model"""

    temporal_unit_range: int = 0
    r"""Maximum temporal unit simulated"""

    z_shape: tuple[int, int] = (0, 0)
    r"""Shape of the Z (intermediate consumption) matrix in the model"""

    y_shape: tuple[int, int] = (0, 0)
    r"""Shape of the Y (final demand) matrix in the model"""

    x_shape: tuple[int, int] = (0, 0)
    r"""Shape of the x (production) vector in the model"""

    regions_idx: npt.NDArray = np.array([])
    r"""lexicographic region indexes"""

    sectors_idx: npt.NDArray = np.array([])
    r"""lexicographic sector indexes"""

    model_monetary_factor: int = 1
    r"""Amount of unitary currency used in the MRIO (e.g. 1000000 if in € millions)"""

    gva_df: pd.Series = pd.Series([], dtype="float64")
    r"""GVA per (region,sector)"""

    sectors_gva_shares: npt.NDArray = np.array([])
    r"""Fraction of total (regional) GVA for each sectors"""

    Z_distrib: npt.NDArray = np.array([])
    r"""Normalized intermediate consumption matrix"""

    Y_distrib: npt.NDArray = np.array([])
    r"""Normalized final consumption matrix"""

    mrio_name: str = ""
    r"""MRIO identification"""

    @abstractmethod
    def __init__(
        self,
        *,
        impact: pd.Series,
        name: str | None = "Unnamed",
        occurrence: int = 1,
        duration: int = 1,
    ) -> None:
        r"""Create an event shocking the model from a dictionary.

        An Event object stores all information about a unique shock during simulation such as
        time of occurrence, duration, type of shock, amount of damages. Computation
        of recovery or initially requested rebuilding demand is also done in this
        class.

        Parameters
        ----------

        event : dict
            A dictionary holding the necessary information to define an event.

        Examples
        --------
            FIXME: Add docs.

        """
        logger.info("Initializing new Event")
        logger.debug("Checking required Class attributes are defined")

        if np.size(self.possible_regions) == 0 or np.size(self.possible_sectors) == 0:
            raise AttributeError(
                "It appears that no model has been instantiated as some class attributes are not initialized (possible_regions, possible_sectors). Events require to instantiate a model and a simulation context before they can be instantiated"
            )

        if self.temporal_unit_range == 0:
            raise AttributeError(
                "It appears that no simulation context has been instantiated as some class attributes are not initialized (temporal_unit_range). Events require to instantiate a model and a simulation context before they can be instantiated"
            )

        self.name: str = name if name else "unnamed"
        r"""An identifying name for the event (for convenience mostly)"""

        self.occurrence = occurrence
        self.duration = duration
        self.impact_df = impact

        self.happened: bool = False
        r"""States if the event happened"""

        self.over: bool = False
        r"""States if the event is over"""

        self.event_dict: dict = {
            "name": str(self.name),
            "occurrence": self.occurrence,
            "duration": self.duration,
            "aff_regions": list(self.aff_regions),
            "aff_sectors": list(self.aff_sectors),
            "impact": self.total_impact,
            "impact_industries_distrib": list(self.impact_industries_distrib),
            "impact_regional_distrib": list(self.impact_regional_distrib),
            "globals_vars": {
                "possible_sectors": list(self.possible_sectors),
                "possible_regions": list(self.possible_regions),
                "temporal_unit_range": self.temporal_unit_range,
                "z_shape": self.z_shape,
                "y_shape": self.y_shape,
                "x_shape": self.x_shape,
                "model_monetary_factor": self.model_monetary_factor,
                "mrio_used": self.mrio_name,
            },
        }
        r"""Store relevant information about the event"""

    @classmethod
    @abstractmethod
    def _instantiate(
        cls,
        impact: pd.Series,
        *,
        occurrence: int = 1,
        duration: int = 1,
        name: Optional[str] = None,
    ):
        return cls(impact=impact, occurrence=occurrence, duration=duration, name=name)

    @classmethod
    def from_series(
        cls,
        impact: pd.Series,
        *,
        occurrence: int = 1,
        duration: int = 1,
        name: Optional[str] = None,
        **kwarg,
    ) -> Event:
        """Create an event from an impact pd.Series.

        Args:
            impact (pd.Series): A vector definition of the impact per (region,sector)
            occurrence (int, optional): The ordinal of occurrence of the event (requires to be > 0). Defaults to 1.
            duration (int, optional): The duration of the event (entire impact applied during this number of steps). Defaults to 1.
            name (Optional[str], optional): A possible name for the event, for convenience. Defaults to None.

        Raises:
            ValueError: Raised if impact is empty or contains negative values.

        Returns:
            Event: an Event object (or one of its subclass).
        """
        if impact.size == 0:
            raise ValueError(
                "Empty impact Series at init, did you not set the impact correctly ?"
            )
        impact = impact[impact != 0]
        if np.less_equal(impact, 0).any():
            logger.debug(f"Impact has negative values:\n{impact}\n{impact[impact<0]}")
            raise ValueError("Impact has negative values")
        return cls._instantiate(
            impact=impact,
            occurrence=occurrence,
            duration=duration,
            name=name,
            **kwarg,
        )

    @classmethod
    def from_dataframe(
        cls,
        impact: pd.DataFrame,
        *,
        occurrence: int = 1,
        duration: int = 1,
        name: Optional[str] = None,
        **kwarg,
    ) -> Event:
        """Convenience function for DataFrames. See :meth:`~boario.event.Event.from_series`

        Raises:
            ValueError: If impact cannot be squeezed to a Series

        Returns:
            Event: an Event object (or one of its subclass).
        """
        impact = impact.squeeze()
        if not isinstance(impact, pd.Series):
            raise ValueError("Could not squeeze impact dataframe to a serie.")

        return cls.from_series(
            impact=impact,
            occurrence=occurrence,
            duration=duration,
            name=name,
            **kwarg,
        )

    @classmethod
    def distribute_impact_by_gva(cls, impact_vec: pd.Series) -> pd.Series:
        """Distribute a vector of impact by the GVA of affected industries.

        Each values of the given impact are mutliplied by the share of the GVA
        the industry has over the GVA of all affected industries.

        Args:
            impact_vec (pd.Series): The impact values to be reweigthed.
            Current use-case assumes all values are the total impact.

        Returns:
            pd.Series: The impact where each value was multiplied by the share
            of GVA of each affected industry (over total GVA affected).
        """
        gva = cls.gva_df.loc[impact_vec.index]
        gva = gva.transform(lambda x: x / sum(x))
        return impact_vec * gva

    @classmethod
    def distribute_impact_equally(cls, impact_vec: pd.Series) -> pd.Series:
        """Distribute an impact equally between all affected regions.

        Assume impact is given as a vector with all value being the
        total impact to distribute.

        Args:
            impact_vec (pd.Series): The impact to distribute.

        Returns:
            pd.Series: The impact vector equally distributed among affected industries.
        """
        dfg = impact_vec.groupby("region")
        return dfg.transform(lambda x: x / (dfg.ngroups * x.count()))

    @classmethod
    def from_scalar_industries(
        cls,
        impact: ScalarImpact,
        *,
        industries: IndustriesList,
        impact_industries_distrib: Optional[npt.ArrayLike] = None,
        gva_distrib: bool = False,
        occurrence: int = 1,
        duration: int = 1,
        name: Optional[str] = None,
        **kwarg,
    ) -> Event:
        """Creates an Event from a scalar and a list of industries affected.

        The scalar impact is distributed evenly by default. Otherwise it can be distributed proportionnaly to the GVA of
        affected industries, or to a custom distribution.

        Args:
            impact (ScalarImpact): The scalar impact.
            industries (IndustriesList): The list of industries affected by the impact
            impact_industries_distrib (Optional[npt.ArrayLike], optional): A vector of equal size to the list of industries, stating the
            share of the impact each industry should receive. Defaults to None.
            gva_distrib (bool, optional): A boolean stating if the impact should be distributed proportionnaly to GVA. Defaults to False.
            occurrence (int, optional): The ordinal of occurrence of the event (requires to be > 0). Defaults to 1.
            duration (int, optional): The duration of the event (entire impact applied during this number of steps). Defaults to 1.
            name (Optional[str], optional): A possible name for the event, for convenience. Defaults to None.

        Raises:
            ValueError: Raise if Impact is null, if len(industries) < 1 or if the sum of impact_industries_distrib differs from 1.0.

        Returns:
            Event: An Event object or one of its subclass.
        """
        if impact <= 0:
            raise ValueError(f"Impact is null")

        if len(industries) < 1:
            raise ValueError(f"Null sized affected industries ?")

        if isinstance(industries, list):
            industries = pd.MultiIndex.from_tuples(
                industries, names=["region", "sector"]
            )

        impact_vec = pd.Series(impact, dtype="float64", index=industries)

        if impact_industries_distrib:
            if np.sum(impact_industries_distrib) != 1.0:
                raise ValueError("Impact distribution doesn't sum up to 1.0")
            else:
                impact_vec *= impact_industries_distrib  # type: ignore

        elif gva_distrib:
            impact_vec = cls.distribute_impact_by_gva(impact_vec)

        else:
            impact_vec = cls.distribute_impact_equally(impact_vec)

        return cls.from_series(
            impact=impact_vec,
            occurrence=occurrence,
            duration=duration,
            name=name,
            **kwarg,
        )

    @classmethod
    def from_scalar_regions_sectors(
        cls,
        impact: ScalarImpact,
        *,
        regions: RegionsList,
        sectors: SectorsList,
        impact_regional_distrib: Optional[npt.ArrayLike] = None,
        impact_sectoral_distrib: Optional[Union[str, npt.ArrayLike]] = None,
        occurrence: int = 1,
        duration: int = 1,
        name: Optional[str] = None,
        **kwarg,
    ) -> Event:
        """Creates an Event from a scalar, a list of regions and a list of sectors affected.



        Args:
            impact (ScalarImpact): The scalar impact.
            regions (RegionsList): The list of regions affected.
            sectors (SectorsList): The list of sectors affected in each region.
            impact_regional_distrib (Optional[npt.ArrayLike], optional): A vector of equal size to the list of regions affected, stating the
            share of the impact each industry should receive. Defaults to None.
            impact_sectoral_distrib (Optional[Union[str, npt.ArrayLike]], optional): A vector of equal size to the list of sectors affected, stating the
            share of the impact each industry should receive. Defaults to None.
            occurrence (int, optional): The ordinal of occurrence of the event (requires to be > 0). Defaults to 1.
            duration (int, optional): The duration of the event (entire impact applied during this number of steps). Defaults to 1.
            name (Optional[str], optional): A possible name for the event, for convenience. Defaults to None.

        Raises:
            ValueError: Raise if Impact is null, if len(regions) or len(sectors) < 1,

        Returns:
            Event: An Event object or one of its subclass
        """
        if impact <= 0:
            raise ValueError(f"Impact is null")

        if isinstance(regions, str):
            regions = [regions]

        if isinstance(sectors, str):
            sectors = [sectors]

        _regions = pd.Index(regions, name="region")
        _sectors = pd.Index(sectors, name="sector")

        if len(_regions) < 1:
            raise ValueError(f"Null sized affected regions ?")

        if len(_sectors) < 1:
            raise ValueError(f"Null sized affected sectors ?")

        if _sectors.duplicated().any():
            warnings.warn(
                UserWarning(
                    "Multiple presence of the same sector in affected sectors. (Will remove duplicate)"
                )
            )
            _sectors = _sectors.drop_duplicates()

        if _regions.duplicated().any():
            warnings.warn(
                UserWarning(
                    "Multiple presence of the same region in affected region. (Will remove duplicate)"
                )
            )
            _regions = _regions.drop_duplicates()

        industries = pd.MultiIndex.from_product(
            [_regions, _sectors], names=["region", "sector"]
        )

        impact_vec = pd.Series(impact, dtype="float64", index=industries)

        assert isinstance(_regions, pd.Index)
        if impact_regional_distrib is None:
            regional_distrib = pd.Series(1.0 / len(_regions), index=_regions)
        elif not isinstance(impact_regional_distrib, pd.Series):
            impact_regional_distrib = pd.Series(impact_regional_distrib, index=_regions)
            regional_distrib = pd.Series(0.0, index=_regions)
            regional_distrib.loc[
                impact_regional_distrib.index
            ] = impact_regional_distrib
        else:
            regional_distrib = pd.Series(0.0, index=_regions)
            try:
                regional_distrib.loc[
                    impact_regional_distrib.index
                ] = impact_regional_distrib
            except KeyError:
                regional_distrib.loc[_regions] = impact_regional_distrib.values

        assert isinstance(_sectors, pd.Index)
        if impact_sectoral_distrib is None:
            sectoral_distrib = pd.Series(1.0 / len(_sectors), index=_sectors)
        elif (
            isinstance(impact_sectoral_distrib, str)
            and impact_sectoral_distrib == "gdp"
        ):
            gva = cls.gva_df.loc[(_regions, _sectors)]
            sectoral_distrib = gva.groupby("region").transform(lambda x: x / sum(x))
        elif not isinstance(impact_sectoral_distrib, pd.Series):
            impact_sectoral_distrib = pd.Series(impact_sectoral_distrib, index=_sectors)
            sectoral_distrib = pd.Series(0.0, index=_sectors)
            sectoral_distrib.loc[
                impact_sectoral_distrib.index
            ] = impact_sectoral_distrib
        else:
            sectoral_distrib = pd.Series(0.0, index=_sectors)
            try:
                sectoral_distrib.loc[
                    impact_sectoral_distrib.index
                ] = impact_sectoral_distrib
            except KeyError:
                sectoral_distrib.loc[_sectors] = impact_sectoral_distrib.values

        logger.debug(f"{sectoral_distrib}")
        logger.debug(f"{regional_distrib}")
        if isinstance(sectoral_distrib.index, pd.MultiIndex):
            industries_distrib = sectoral_distrib * regional_distrib
        else:
            industries_distrib = pd.Series(
                np.outer(regional_distrib.values, sectoral_distrib.values).flatten(),  # type: ignore
                index=pd.MultiIndex.from_product(
                    [regional_distrib.index, sectoral_distrib.index]
                ),
            )

        impact_vec *= industries_distrib

        return cls.from_series(
            impact=impact_vec,
            occurrence=occurrence,
            duration=duration,
            name=name,
            **kwarg,
        )

    @property
    def impact_df(self) -> pd.Series:
        r"""A pandas Series with all possible industries as index, holding the impact vector of the event. The impact is defined for each sectors in each region."""
        return self._impact_df

    @impact_df.setter
    def impact_df(self, value: pd.Series):
        self._impact_df = pd.Series(
            0,
            dtype="float64",
            index=pd.MultiIndex.from_product(
                [self.possible_regions, self.possible_sectors],
                names=["region", "sector"],
            ),
        )
        self._impact_df[value.index] = value
        logger.debug("Sorting impact Series")
        self._impact_df.sort_index(inplace=True)
        self.aff_industries = self.impact_df.loc[self.impact_df > 0].index  # type: ignore
        self.impact_industries_distrib = self.impact_df.transform(lambda x: x / sum(x))
        self.total_impact = self.impact_df.sum()
        self.impact_vector = self.impact_df.values

    @property
    def aff_industries(self) -> pd.MultiIndex:
        r"""The industries affected by the event.

        Parameters
        ----------

        index : pd.MultiIndex
             The affected industries as a pandas MultiIndex

        Returns
        -------
             A pandas MultiIndex with the regions affected as first level, and sectors affected as second level

        """

        return self._aff_industries

    @aff_industries.setter
    def aff_industries(self, index: pd.MultiIndex):
        if not isinstance(index, pd.MultiIndex):
            raise ValueError(
                "Given impact vector does not have a (region,sector) MultiIndex"
            )
        if index.names != ["region", "sector"]:
            raise ValueError("MultiIndex must have levels named 'region' and 'sector'")

        regions = index.get_level_values("region").unique()
        sectors = index.get_level_values("sector").unique()
        if not set(regions).issubset(self.possible_regions):
            raise ValueError(
                f"Regions {set(regions) - set(self.possible_regions)} are not in the possible regions list"
            )
        if not set(sectors).issubset(self.possible_sectors):
            raise ValueError(
                f"Sectors {set(sectors) - set(self.possible_sectors)} are not in the possible sectors list"
            )

        self.aff_regions = regions
        self.aff_sectors = sectors
        logger.debug(
            f"Setting _aff_industries. There are {np.size(index)} affected industries"
        )
        self._aff_industries = index
        self._aff_industries_idx = np.array(
            [
                np.size(self.possible_sectors) * ri + si
                for ri in self._aff_regions_idx
                for si in self._aff_sectors_idx
            ]
        )

    @property
    def occurrence(self) -> int:
        r"""The temporal unit of occurrence of the event."""

        return self._occur

    @occurrence.setter
    def occurrence(self, value: int):
        if not 0 < value <= self.temporal_unit_range:
            raise ValueError(
                "Occurrence of event is not in the range of simulation steps (cannot be 0) : {} not in ]0-{}]".format(
                    value, self.temporal_unit_range
                )
            )
        else:
            logger.debug(f"Setting occurence to {value}")
            self._occur = value

    @property
    def duration(self) -> int:
        r"""The duration of the event."""

        return self._duration

    @duration.setter
    def duration(self, value: int):
        if not 0 <= self.occurrence + value <= self.temporal_unit_range:
            raise ValueError(
                "Occurrence + duration of event is not in the range of simulation steps : {} + {} not in [0-{}]".format(
                    self.occurrence, value, self.temporal_unit_range
                )
            )
        else:
            logger.debug(f"Setting duration to {value}")
            self._duration = value

    @property
    def aff_regions(self) -> pd.Index:
        r"""The array of regions affected by the event"""

        return self._aff_regions

    @property
    def aff_regions_idx(self) -> npt.NDArray:
        r"""The array of lexicographically ordered affected region indexes"""

        return self._aff_regions_idx

    @aff_regions.setter
    def aff_regions(self, value: npt.ArrayLike | str):
        if isinstance(value, str):
            value = [value]
        value = pd.Index(value)  # type: ignore
        impossible_regions = np.setdiff1d(value, self.possible_regions)
        if impossible_regions.size > 0:
            raise ValueError(
                "These regions are not in the model : {}".format(impossible_regions)
            )
        else:
            self._aff_regions = pd.Index(value, name="region")
            logger.debug(f"Setting _aff_regions to {self._aff_regions}")
            self._aff_regions_idx = np.searchsorted(self.possible_regions, value)

    @property
    def aff_sectors(self) -> pd.Index:
        r"""The array of affected sectors by the event"""

        return self._aff_sectors

    @property
    def aff_sectors_idx(self) -> npt.NDArray:
        r"""The array of lexicographically ordered affected sectors indexes"""

        return self._aff_sectors_idx

    @aff_sectors.setter
    def aff_sectors(self, value: npt.ArrayLike | str):
        if isinstance(value, str):
            value = [value]
        value = pd.Index(value, name="sector")  # type: ignore
        impossible_sectors = np.setdiff1d(value, self.possible_sectors)
        if impossible_sectors.size > 0:
            raise ValueError(
                "These sectors are not in the model : {}".format(impossible_sectors)
            )
        else:
            self._aff_sectors = pd.Index(value, name="sector")
            logger.debug(f"Setting _aff_sectors to {self._aff_sectors}")
            self._aff_sectors_idx = np.searchsorted(self.possible_sectors, value)

    @property
    def impact_regional_distrib(self) -> pd.Series:
        r"""The series specifying how damages are distributed among affected regions"""

        return self._impact_regional_distrib

    @property
    def impact_industries_distrib(self) -> pd.Series:
        r"""The series specifying how damages are distributed among affected industries (regions,sectors)"""

        return self._impact_industries_distrib

    @impact_industries_distrib.setter
    def impact_industries_distrib(self, value: pd.Series):
        self._impact_industries_distrib = value
        self._impact_regional_distrib = self._impact_industries_distrib.groupby(
            "region"
        ).sum()

    def __repr__(self):
        # TODO: find ways to represent long lists
        return f"""
        Event(
              name = {self.name},
              occur = {self.occurrence},
              duration = {self.duration}
              aff_regions = {self.aff_regions.to_list()},
              aff_sectors = {self.aff_sectors.to_list()},
             )
        """


class EventArbitraryProd(Event):
    def __init__(
        self,
        *,
        impact: pd.Series,
        recovery_time: int = 1,
        recovery_function: str = "linear",
        name: str | None = "Unnamed",
        occurrence: int = 1,
        duration: int = 1,
    ) -> None:
        if (impact > 1.0).any():
            raise ValueError(
                "Impact is greater than 100% (1.) for at least an industry."
            )

        super().__init__(
            impact=impact,
            name=name,
            occurrence=occurrence,
            duration=duration,
        )

        self._prod_cap_delta_arbitrary_0 = (
            self.impact_vector.copy()
        )  # np.zeros(shape=len(self.possible_sectors))
        self.prod_cap_delta_arbitrary = (
            self.impact_vector.copy()
        )  # type: ignore # np.zeros(shape=len(self.possible_sectors))
        self.recovery_time = recovery_time
        r"""The characteristic recovery duration after the event is over"""

        self.recovery_function = recovery_function

        logger.info("Initialized")

    @classmethod
    def _instantiate(
        cls,
        impact: pd.Series,
        *,
        recovery_time: int = 1,
        recovery_function: str = "linear",
        occurrence: int = 1,
        duration: int = 1,
        name: Optional[str] = None,
    ):
        return cls(
            impact=impact,
            occurrence=occurrence,
            duration=duration,
            name=name,
            recovery_time=recovery_time,
            recovery_function=recovery_function,
        )

    @property
    def prod_cap_delta_arbitrary(self) -> npt.NDArray:
        r"""The optional array storing arbitrary (as in not related to productive_capital destroyed) production capacity loss"""
        return self._prod_cap_delta_arbitrary

    @prod_cap_delta_arbitrary.setter
    def prod_cap_delta_arbitrary(self, value: npt.NDArray):
        self._prod_cap_delta_arbitrary = value

    @property
    def recoverable(self) -> bool:
        r"""A boolean stating if an event can start recover"""
        return self._recoverable

    @recoverable.setter
    def recoverable(self, current_temporal_unit: int):
        reb = (self.occurrence + self.duration) <= current_temporal_unit
        if reb and not self.recoverable:
            logger.info(
                "Temporal_Unit : {} ~ Event named {} that occured at {} in {} has started recovering (arbitrary production capacity loss)".format(
                    current_temporal_unit,
                    self.name,
                    self.occurrence,
                    self.aff_regions.to_list(),
                )
            )
        self._recoverable = reb

    @property
    def recovery_function(self) -> Callable:
        r"""The recovery function used for recovery (`Callable`)"""
        return self._recovery_fun

    @recovery_function.setter
    def recovery_function(self, r_fun: str | Callable | None):
        if r_fun is None:
            r_fun = "instant"
        if self.recovery_time is None:
            raise AttributeError(
                "Impossible to set recovery function if no recovery time is given."
            )
        if isinstance(r_fun, str):
            if r_fun == "linear":
                fun = linear_recovery
            elif r_fun == "convexe":
                fun = convexe_recovery_scaled
            elif r_fun == "convexe noscale":
                fun = convexe_recovery
            elif r_fun == "concave":
                fun = concave_recovery
            else:
                raise NotImplementedError(
                    "No implemented recovery function corresponding to {}".format(r_fun)
                )
        elif callable(r_fun):
            r_fun_argsspec = inspect.getfullargspec(r_fun)
            r_fun_args = r_fun_argsspec.args + r_fun_argsspec.kwonlyargs
            if not all(
                args in r_fun_args
                for args in [
                    "init_impact_stock",
                    "elapsed_temporal_unit",
                    "recovery_time",
                ]
            ):
                raise ValueError(
                    "Recovery function has to have at least the following keyword arguments: {}".format(
                        [
                            "init_impact_stock",
                            "elapsed_temporal_unit",
                            "recovery_time",
                        ]
                    )
                )
            fun = r_fun

        else:
            raise ValueError("Given recovery function is not a str or callable")

        r_fun_partial = partial(
            fun,
            init_impact_stock=self._prod_cap_delta_arbitrary_0,
            recovery_time=self.recovery_time,
        )
        self._recovery_fun = r_fun_partial

    def recovery(self, current_temporal_unit: int):
        elapsed = current_temporal_unit - (self.occurrence + self.duration)
        if elapsed < 0:
            raise RuntimeError("Trying to recover before event is over")
        if self.recovery_function is None:
            raise RuntimeError(
                "Trying to recover event while recovery function isn't set yet"
            )
        res = self.recovery_function(elapsed_temporal_unit=elapsed)
        precision = int(math.log10(self.model_monetary_factor)) + 1
        res = np.around(res, precision)
        if not np.any(res):
            self.over = True
        self._prod_cap_delta_arbitrary = res


class EventKapitalDestroyed(Event, ABC):
    def __init__(
        self,
        *,
        impact: pd.Series,
        households_impact: Optional[Impact] = None,
        name: str | None = "Unnamed",
        occurrence: int = 1,
        duration: int = 1,
        event_monetary_factor: Optional[int] = None,
    ) -> None:
        if event_monetary_factor is None:
            logger.info(
                f"No event monetary factor given. Assuming it is the same as the model ({self.model_monetary_factor})"
            )
            self.event_monetary_factor = self.model_monetary_factor
            r"""The monetary factor for the impact of the event (e.g. 10**6, 10**3, ...)"""

        else:
            self.event_monetary_factor = event_monetary_factor
            if self.event_monetary_factor != self.model_monetary_factor:
                logger.warning(
                    f"Event monetary factor ({self.event_monetary_factor}) differs from model monetary factor ({self.model_monetary_factor}). Be careful to define your impact with the correct unit (ie in event monetary factor)."
                )

        if (impact < LOW_DEMAND_THRESH / self.event_monetary_factor).all():
            raise ValueError(
                "Impact is too small to be considered by the model. Check you units perhaps ?"
            )

        super().__init__(
            impact=impact,
            name=name,
            occurrence=occurrence,
            duration=duration,
        )

        # The only thing we have to do is affecting/computing the regional_sectoral_productive_capital_destroyed
        self.impact_df = self.impact_df * (
            self.event_monetary_factor / self.model_monetary_factor
        )
        self.total_productive_capital_destroyed = self.total_impact
        logger.info(
            f"Total impact on productive capital is {self.total_productive_capital_destroyed} (in model unit)"
        )
        self.remaining_productive_capital_destroyed = (
            self.total_productive_capital_destroyed
        )
        self._regional_sectoral_productive_capital_destroyed_0 = (
            self.impact_vector.copy()
        )
        self.regional_sectoral_productive_capital_destroyed = self.impact_vector.copy()  # type: ignore
        self.households_impact_df: pd.Series = pd.Series(
            0,
            dtype="float64",
            index=pd.MultiIndex.from_product(
                [self.possible_regions, self.possible_final_demand_cat],
                names=["region", "final_demand_cat"],
            ),
        )
        r"""A pandas Series with all possible (regions, final_demand_cat) as index, holding the households impacts vector of the event. The impact is defined for each region and each final demand category."""

        if households_impact:
            try:
                rebuilding_demand_idx = self.possible_final_demand_cat[
                    self.possible_final_demand_cat.str.match(
                        REBUILDING_FINALDEMAND_CAT_REGEX
                    )
                ]  # .values[0]
                if len(rebuilding_demand_idx) > 1:
                    raise ValueError(
                        f"The rebuilding demand index ({rebuilding_demand_idx}) contains multiple values which is a problem. Contact the dev to update the matching regexp."
                    )

            except IndexError:
                logger.warning(
                    f"No final demand category matched common rebuilding final demand category, hence we will put it in the first available ({self.possible_final_demand_cat[0]})."
                )
                rebuilding_demand_idx = self.possible_final_demand_cat[0]
            if isinstance(households_impact, pd.Series):
                logger.debug("Given household impact is a pandas Series")
                self.households_impact_df.loc[
                    households_impact.index, rebuilding_demand_idx
                ] = households_impact  # type: ignore
            elif isinstance(households_impact, dict):
                logger.debug("Given household impact is a dict")
                households_impact = pd.Series(households_impact, dtype="float64")
                self.households_impact_df.loc[
                    households_impact.index, rebuilding_demand_idx
                ] = households_impact  # type: ignore
            elif isinstance(households_impact, pd.DataFrame):
                logger.debug("Given household impact is a dataframe")
                households_impact = households_impact.squeeze()
                if not isinstance(households_impact, pd.Series):
                    raise ValueError(
                        "The given households_impact DataFrame is not a Series after being squeezed"
                    )
                self.households_impact_df.loc[
                    households_impact.index, rebuilding_demand_idx
                ] = households_impact  # type: ignore
            elif isinstance(households_impact, (list, np.ndarray)):
                if np.size(households_impact) != np.size(self.aff_regions):
                    raise ValueError(
                        f"Length mismatch between households_impact ({np.size(households_impact)} and affected regions ({np.size(self.aff_regions)}))"
                    )
                else:
                    self.households_impact_df.loc[
                        self.aff_regions, rebuilding_demand_idx
                    ] = households_impact  # type: ignore
            elif isinstance(households_impact, (float, int)):
                if self.impact_regional_distrib is not None:
                    logger.warning(
                        f"households impact ({households_impact}) given as scalar, distributing among region following `impact_regional_distrib` ({self.impact_regional_distrib}) to {self.aff_regions}"
                    )
                    logger.debug(f"{rebuilding_demand_idx}")
                    self.households_impact_df.loc[
                        self.aff_regions, rebuilding_demand_idx
                    ] = (
                        households_impact * self.impact_regional_distrib
                    )  # type: ignore
            self.households_impact_df *= (
                self.event_monetary_factor / self.model_monetary_factor
            )
        logger.info(
            f"Total impact on households is {self.households_impact_df.sum()} (in model unit)"
        )

    @property
    def regional_sectoral_productive_capital_destroyed(self) -> npt.NDArray:
        r"""The (optional) array of productive_capital destroyed per industry (ie region x sector)"""
        return self._regional_sectoral_productive_capital_destroyed

    @regional_sectoral_productive_capital_destroyed.setter
    def regional_sectoral_productive_capital_destroyed(self, value: npt.ArrayLike):
        value = np.array(value)
        logger.debug(
            f"Changing regional_sectoral_productive_capital_destroyed with {value}\n Sum is {value.sum()}"
        )
        if value.shape != self.x_shape:
            raise ValueError(
                "Incorrect shape give for regional_sectoral_productive_capital_destroyed: {} given and {} expected".format(
                    value.shape, self.x_shape
                )
            )
        self._regional_sectoral_productive_capital_destroyed = value


class EventKapitalRebuild(EventKapitalDestroyed):
    def __init__(
        self,
        *,
        impact: pd.Series,
        households_impact: Impact | None = None,
        name: str | None = "Unnamed",
        occurrence: int = 1,
        duration: int = 1,
        event_monetary_factor: Optional[int] = None,
        rebuild_tau: int,
        rebuilding_sectors: dict[str, float] | pd.Series,
        rebuilding_factor: float = 1.0,
    ) -> None:
        super().__init__(
            impact=impact,
            households_impact=households_impact,
            name=name,
            occurrence=occurrence,
            duration=duration,
            event_monetary_factor=event_monetary_factor,
        )
        self._rebuildable = False
        self.rebuild_tau = rebuild_tau
        self.rebuilding_sectors = rebuilding_sectors
        self.rebuilding_factor = rebuilding_factor

        industrial_rebuilding_demand = np.zeros(shape=self.z_shape)
        tmp = np.zeros(self.z_shape, dtype="float")
        mask = np.ix_(
            np.union1d(
                self._rebuilding_industries_RoW_idx, self._rebuilding_industries_idx
            ),
            self._aff_industries_idx,
        )
        industrial_rebuilding_demand = np.outer(
            self.rebuilding_sectors_shares,
            self.regional_sectoral_productive_capital_destroyed,
        )
        industrial_rebuilding_demand = industrial_rebuilding_demand * rebuilding_factor
        tmp[mask] = self.Z_distrib[mask] * industrial_rebuilding_demand[mask]

        precision = int(math.log10(self.model_monetary_factor)) + 1
        np.around(tmp, precision, tmp)

        self.rebuilding_demand_indus = tmp

        households_rebuilding_demand = np.zeros(shape=self.y_shape)
        tmp = np.zeros(self.y_shape, dtype="float")
        mask = np.ix_(  # type: ignore
            np.union1d(  # type: ignore
                self._rebuilding_industries_RoW_idx, self._rebuilding_industries_idx
            ),
            list(range(self.y_shape[1])),
        )

        households_rebuilding_demand = np.outer(
            self.rebuilding_sectors_shares, self.households_impact_df.to_numpy()
        )
        households_rebuilding_demand = households_rebuilding_demand * rebuilding_factor
        tmp[mask] = self.Y_distrib[mask] * households_rebuilding_demand[mask]
        np.around(tmp, precision, tmp)
        self.rebuilding_demand_house = tmp
        logger.debug(
            f"house rebuilding demand is: {pd.DataFrame(self.rebuilding_demand_house, index=pd.MultiIndex.from_product([self.possible_regions,self.possible_sectors]))}"
        )

        self.event_dict["rebuilding_sectors"] = {
            sec: share
            for sec, share in zip(
                self.rebuilding_sectors, self.rebuilding_sectors_shares
            )
        }

    @classmethod
    def _instantiate(
        cls,
        impact: pd.Series,
        *,
        households_impact: Optional[Impact] = None,
        occurrence: int = 1,
        duration: int = 1,
        name: str | None = None,
        event_monetary_factor: Optional[int] = None,
        rebuild_tau: int,
        rebuilding_sectors: dict[str, float] | pd.Series,
        rebuilding_factor: float = 1.0,
    ):
        return cls(
            impact=impact,
            households_impact=households_impact,
            occurrence=occurrence,
            duration=duration,
            name=name,
            event_monetary_factor=event_monetary_factor,
            rebuild_tau=rebuild_tau,
            rebuilding_sectors=rebuilding_sectors,
            rebuilding_factor=rebuilding_factor,
        )

    @property
    def rebuilding_sectors(self) -> pd.Index:
        r"""The (optional) array of rebuilding sectors"""
        return self._rebuilding_sectors

    @property
    def rebuilding_sectors_idx(self) -> npt.NDArray:
        r"""The (optional) array of indexes of the rebuilding sectors (in lexicographic order)"""
        return self._rebuilding_sectors_idx

    @property
    def rebuilding_sectors_shares(self) -> npt.NDArray:
        r"""The array specifying how rebuilding demand is distributed along the rebuilding sectors"""
        return self._rebuilding_sectors_shares

    @rebuilding_sectors.setter
    def rebuilding_sectors(self, value: dict[str, float] | pd.Series):
        if isinstance(value, dict):
            reb_sectors = pd.Series(value)
        else:
            reb_sectors = value
        assert np.isclose(reb_sectors.sum(), 1.0)
        impossible_sectors = np.setdiff1d(reb_sectors.index, self.possible_sectors)
        if impossible_sectors.size > 0:
            raise ValueError(
                "These sectors are not in the model : {}".format(impossible_sectors)
            )
        else:
            self._rebuilding_sectors = reb_sectors.index
            self._rebuilding_sectors_idx = np.searchsorted(
                self.possible_sectors, reb_sectors.index
            )

            self._rebuilding_sectors_shares = np.zeros(self.x_shape)
            self._rebuilding_industries_idx = np.array(
                [
                    np.size(self.possible_sectors) * ri + si
                    for ri in self._aff_regions_idx
                    for si in self._rebuilding_sectors_idx
                ]
            )
            self._rebuilding_industries_RoW_idx = np.array(
                [
                    np.size(self.possible_sectors) * ri + si
                    for ri in range(np.size(self.possible_regions))
                    if ri not in self._aff_regions_idx
                    for si in self._rebuilding_sectors_idx
                ]
            )
            self._rebuilding_sectors_shares[self._rebuilding_industries_idx] = np.tile(
                np.array(reb_sectors.values), np.size(self.aff_regions)
            )
            self._rebuilding_sectors_shares[
                self._rebuilding_industries_RoW_idx
            ] = np.tile(
                np.array(reb_sectors.values),
                (np.size(self.possible_regions) - np.size(self.aff_regions)),
            )

    @property
    def rebuilding_demand_house(self) -> npt.NDArray:
        r"""The optional array of rebuilding demand from households"""
        return self._rebuilding_demand_house

    @rebuilding_demand_house.setter
    def rebuilding_demand_house(self, value: npt.ArrayLike):
        value = np.array(value)
        if value.shape != self.y_shape:
            raise ValueError(
                "Incorrect shape give for rebuilding_demand_house: {} given and {} expected".format(
                    value.shape, self.y_shape
                )
            )
        value[value < LOW_DEMAND_THRESH / self.model_monetary_factor] = 0.0
        self._rebuilding_demand_house = value

    @property
    def rebuilding_demand_indus(self) -> npt.NDArray:
        r"""The optional array of rebuilding demand from industries (to rebuild their productive_capital)"""
        return self._rebuilding_demand_indus

    @rebuilding_demand_indus.setter
    def rebuilding_demand_indus(self, value: npt.ArrayLike):
        value = np.array(value)
        if value.shape != self.z_shape:
            raise ValueError(
                "Incorrect shape give for rebuilding_demand_indus: {} given and {} expected".format(
                    value.shape, self.z_shape
                )
            )
        value[value < LOW_DEMAND_THRESH / self.model_monetary_factor] = 0.0
        self._rebuilding_demand_indus = value
        # Also update productive_capital destroyed
        self.regional_sectoral_productive_capital_destroyed = value.sum(axis=0) * (
            1 / self.rebuilding_factor
        )

    @property
    def rebuildable(self) -> bool:
        r"""A boolean stating if an event can start rebuild"""
        return self._rebuildable

    @rebuildable.setter
    def rebuildable(self, current_temporal_unit: int):
        reb = (self.occurrence + self.duration) <= current_temporal_unit
        if reb and not self.rebuildable:
            logger.info(
                "Temporal_Unit : {} ~ Event named {} that occured at {} in {} for {} damages has started rebuilding".format(
                    current_temporal_unit,
                    self.name,
                    self.occurrence,
                    self.aff_regions.to_list(),
                    self.total_productive_capital_destroyed,
                )
            )
        self._rebuildable = reb


class EventKapitalRecover(EventKapitalDestroyed):
    def __init__(
        self,
        *,
        impact: pd.Series,
        recovery_time: int,
        recovery_function: str = "linear",
        households_impact: Optional[Impact] = None,
        name: str | None = "Unnamed",
        occurrence=1,
        duration=1,
        event_monetary_factor=None,
    ) -> None:
        super().__init__(
            impact=impact,
            households_impact=households_impact,
            name=name,
            occurrence=occurrence,
            duration=duration,
            event_monetary_factor=event_monetary_factor,
        )
        self.recovery_time = recovery_time
        self.recovery_function = recovery_function
        self._recoverable = False

    @classmethod
    def _instantiate(
        cls,
        impact: pd.Series,
        *,
        households_impact: Optional[Impact] = None,
        occurrence: int = 1,
        duration: int = 1,
        name: str | None = None,
        event_monetary_factor: Optional[int] = None,
        recovery_time: int,
        recovery_function: str = "linear",
    ):
        return cls(
            impact=impact,
            households_impact=households_impact,
            occurrence=occurrence,
            duration=duration,
            name=name,
            event_monetary_factor=event_monetary_factor,
            recovery_time=recovery_time,
            recovery_function=recovery_function,
        )

    @property
    def recoverable(self) -> bool:
        r"""A boolean stating if an event can start recover"""
        return self._recoverable

    @recoverable.setter
    def recoverable(self, current_temporal_unit: int):
        reb = (self.occurrence + self.duration) <= current_temporal_unit
        if reb and not self.recoverable:
            logger.info(
                "Temporal_Unit : {} ~ Event named {} that occured at {} in {} for {} damages has started recovering (no rebuilding demand)".format(
                    current_temporal_unit,
                    self.name,
                    self.occurrence,
                    self.aff_regions.to_list(),
                    self.total_productive_capital_destroyed,
                )
            )
        self._recoverable = reb

    @property
    def recovery_function(self) -> Callable:
        r"""The recovery function used for recovery (`Callable`)"""
        return self._recovery_fun

    @recovery_function.setter
    def recovery_function(self, r_fun: str | Callable | None):
        if r_fun is None:
            r_fun = "linear"
        if self.recovery_time is None:
            raise AttributeError(
                "Impossible to set recovery function if no recovery time is given."
            )
        if isinstance(r_fun, str):
            if r_fun == "linear":
                fun = linear_recovery
            elif r_fun == "convexe":
                fun = convexe_recovery_scaled
            elif r_fun == "convexe noscale":
                fun = convexe_recovery
            elif r_fun == "concave":
                fun = concave_recovery
            else:
                raise NotImplementedError(
                    "No implemented recovery function corresponding to {}".format(r_fun)
                )
        elif callable(r_fun):
            r_fun_argsspec = inspect.getfullargspec(r_fun)
            r_fun_args = r_fun_argsspec.args + r_fun_argsspec.kwonlyargs
            if not all(
                args in r_fun_args
                for args in [
                    "init_impact_stock",
                    "elapsed_temporal_unit",
                    "recovery_time",
                ]
            ):
                raise ValueError(
                    "Recovery function has to have at least the following keyword arguments: {}".format(
                        [
                            "init_impact_stock",
                            "elapsed_temporal_unit",
                            "recovery_time",
                        ]
                    )
                )
            fun = r_fun

        else:
            raise ValueError("Given recovery function is not a str or callable")

        r_fun_partial = partial(
            fun,
            init_impact_stock=self._regional_sectoral_productive_capital_destroyed_0,
            recovery_time=self.recovery_time,
        )
        self._recovery_fun = r_fun_partial

    def recovery(self, current_temporal_unit: int):
        elapsed = current_temporal_unit - (self.occurrence + self.duration)
        if elapsed < 0:
            raise RuntimeError("Trying to recover before event is over")
        if self.recovery_function is None:
            raise RuntimeError(
                "Trying to recover event while recovery function isn't set yet"
            )
        res = self.recovery_function(elapsed_temporal_unit=elapsed)
        precision = int(math.log10(self.model_monetary_factor)) + 1
        res = np.around(res, precision)
        res[res < 0] = 0.0
        if not np.any(res):
            self.over = True
        self.regional_sectoral_productive_capital_destroyed = res



# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import json
import pathlib
import typing
from typing import Optional
import numpy as np
import numpy.typing as npt
import pandas as pd
from pymrio.core.mriosystem import IOSystem

from boario import logger, warn_once
from boario.event import (
    Event,
    EventKapitalDestroyed,
    EventKapitalRebuild,
    EventArbitraryProd,
)
from boario.utils.misc import lexico_reindex


__all__ = [
    "ARIOBaseModel",
    "INV_THRESHOLD",
    "VALUE_ADDED_NAMES",
    "VA_idx",
    "lexico_reindex",
]

INV_THRESHOLD = 0  # 20 #temporal_units

TECHNOLOGY_THRESHOLD = (
    0.00001  # Do not care about input if producing requires less than this value
)

VALUE_ADDED_NAMES = [
    "VA",
    "Value Added",
    "value added",
    "factor inputs",
    "factor_inputs",
    "Factors Inputs",
    "Satellite Accounts",
    "satellite accounts",
    "satellite_accounts",
    "satellite",
]

VA_idx = np.array(
    [
        "Taxes less subsidies on products purchased: Total",
        "Other net taxes on production",
        "Compensation of employees; wages, salaries, & employers' social contributions: Low-skilled",
        "Compensation of employees; wages, salaries, & employers' social contributions: Medium-skilled",
        "Compensation of employees; wages, salaries, & employers' social contributions: High-skilled",
        "Operating surplus: Consumption of fixed capital",
        "Operating surplus: Rents on land",
        "Operating surplus: Royalties on resources",
        "Operating surplus: Remaining net operating surplus",
    ],
    dtype=object,
)


class ARIOBaseModel:
    def __init__(
        self,
        pym_mrio: IOSystem,
        *,
        order_type: str = "alt",
        alpha_base: float = 1.0,
        alpha_max: float = 1.25,
        alpha_tau: int = 365,
        rebuild_tau: int = 60,
        main_inv_dur: int = 90,
        monetary_factor: int = 10**6,
        temporal_units_by_step: int = 1,
        iotable_year_to_temporal_unit_factor: int = 365,
        infinite_inventories_sect: Optional[list] = None,
        inventory_dict: Optional[dict] = None,
        productive_capital_vector: Optional[
            pd.Series | npt.NDArray | pd.DataFrame
        ] = None,
        productive_capital_to_VA_dict: Optional[dict] = None,
    ) -> None:
        r"""The core of an ARIO model.  Handles the different arrays containing the mrio tables.

        `ARIOBaseModel` wrap all the data and functions used in the core of the most basic version of the ARIO
        model (based on :cite:`2013:hallegatte` and :cite:`2020:guan`).

        An ARIOBaseModel instance based is build on the given IOSystem.
        Most default parameters are the same as in :cite:`2013:hallegatte` and
        default order mechanisms comes from :cite:`2020:guan`. By default each
        industry capital stock is 4 times its value added (:cite:`2008:hallegatte`).

        Parameters
        ----------
        pym_mrio : IOSystem
            The IOSystem to base the model on.
        order_type : str, default "alt"
            The type of orders mechanism to use. Currently, can be "alt" or
            "noalt". See :ref:`boario-math`
        alpha_base : float, default 1.0
            Base value of overproduction factor :math:`\alpha^{b}` (Default to 1.0).
        alpha_max : float, default 1.25
            Maximum factor of overproduction :math:`\alpha^{\textrm{max}}` (default should be 1.25).
        alpha_tau : int, default 365
            Characteristic time of overproduction :math:`\tau_{\alpha}` in ``n_temporal_units_by_step`` (default should be 365 days).
        rebuild_tau : int, default 60
            Rebuilding characteristic time :math:`\tau_{\textrm{REBUILD}}` (see :ref:`boario-math`). Overwritten by per event value if it exists.
        main_inv_dur : int, default 90
            The default numbers of days for inputs inventory to use if it is not defined for an input.
        monetary_factor : int, default 10**6
            Monetary unit factor (i.e. if the tables unit is 10^6 € instead of 1 €, it should be set to 10^6).
        temporal_units_by_step: int, default 1
            The number of temporal_units between each step. (Current version of the model showed inconsistencies with values other than `1`).
        iotable_year_to_temporal_unit_factor : int, default 365
            The (inverse of the) factor by which MRIO data should be multiplied in order to get "per temporal unit value", assuming IO data is yearly.
        infinite_inventories_sect: list, optional
            List of inputs (sector) considered never constraining for production.
        inventory_dict: dict, optional
            Dictionary in the form input:initial_inventory_size, (where input is a sector, and initial_inventory_size is in "temporal_unit" (defaults to a day))
        productive_capital_vector: ArrayLike, optional
            Array of capital per industry if you need to give it exogenously.
        productive_capital_to_VA_dict: dict, optional
            Dictionary in the form sector:ratio. Where ratio is used to estimate capital stock based on the value added of the sector.

        Notes
        -----

        It is recommended to use ``productive_capital_to_VA_dict`` if you have a more precise estimation of
        the ratio of (Capital Stock / Value Added) per sectors than the default 4/1 ratio.
        You may also feed in directly a ``productive_capital_vector`` if you did your estimation before-hand.
        (This is especially useful if you have events based of an exposure layer for instance)

        Regarding inventories, they default to 90 days for all inputs (ie sectors).
        You can set some inputs to be never constraining for production by listing them
        in ``infinite_inventories_sect`` or directly feed in a dictionary of the inventory
        duration for each input.

        """

        logger.debug("Initiating new ARIOBaseModel instance")
        super().__init__()
        ################ Parameters variables #######################
        try:
            logger.info(
                "IO system metadata :\n{}\n{}\n{}\n{}".format(
                    str(pym_mrio.meta.description),
                    str(pym_mrio.meta.name),
                    str(pym_mrio.meta.system),
                    str(pym_mrio.meta.version),
                )
            )
        except AttributeError:
            logger.warning(
                "It seems the MRIOT you loaded doesn't have metadata to print."
            )

        pym_mrio = lexico_reindex(pym_mrio)
        self.main_inv_dur: int = main_inv_dur
        r"""int, default 90: The default numbers of days for inputs inventory to use if it is not defined for an input."""

        reg = pym_mrio.get_regions()
        reg = typing.cast("list[str]", reg)
        self.regions = np.array(sorted(reg))
        r"""numpy.ndarray of str : An array of the regions of the model."""

        self.n_regions = len(reg)
        r"""int : The number :math:`m` of regions."""

        sec = pym_mrio.get_sectors()
        sec = typing.cast("list[str]", sec)
        self.sectors = np.array(sorted(sec))
        r"""numpy.ndarray of str: An array of the sectors of the model."""

        self.n_sectors = len(sec)
        r"""int : The number :math:`n` of sectors."""

        self.industries = pd.MultiIndex.from_product(
            [self.regions, self.sectors], names=["region", "sector"]
        )
        r"""pandas.MultiIndex : A pandas MultiIndex of the industries (region,sector) of the model."""

        self.n_industries = len(self.industries)
        r"""int : The number :math:`m * n` of industries."""

        try:
            self.final_demand_cat = np.array(sorted(list(pym_mrio.get_Y_categories())))  # type: ignore
            r"""numpy.ndarray of str: An array of the final demand categories of the model (``["Final demand"]`` if there is only one)"""

            self.n_fd_cat = len(pym_mrio.get_Y_categories())  # type: ignore
            r"""int: The numbers of final demand categories."""

        except KeyError:
            self.n_fd_cat = 1
            self.final_demand_cat = np.array(["Final demand"])
        except IndexError:
            self.n_fd_cat = 1
            self.final_demand_cat = np.array(["Final demand"])

        if hasattr(pym_mrio, "monetary_factor"):
            logger.warning(
                f"Custom monetary factor found in the mrio pickle file, continuing with this one ({getattr(pym_mrio,'monetary_factor')})"
            )
            self.monetary_factor = getattr(pym_mrio, "monetary_factor")
        else:
            self.monetary_factor = monetary_factor
        r"""int, default 10^6: Monetary unit factor (i.e. if the tables unit is 10^6 € instead of 1 €, it should be set to 10^6)."""

        self.n_temporal_units_by_step = temporal_units_by_step
        r"""int, default 1: The number of temporal_units between each step. (Current version of the model was not tested with values other than `1`)."""

        self.iotable_year_to_temporal_unit_factor = iotable_year_to_temporal_unit_factor
        r"""int, default 365: The (inverse of the) factor by which MRIO data should be multiplied in order to get "per temporal unit value", assuming IO data is yearly."""

        if self.iotable_year_to_temporal_unit_factor != 365:
            logger.warning(
                "iotable_to_daily_step_factor is not set to 365 (days). This should probably not be the case if the IO tables you use are on a yearly basis."
            )

        self.steply_factor = (
            self.n_temporal_units_by_step / self.iotable_year_to_temporal_unit_factor
        )
        self.rebuild_tau = rebuild_tau
        r"""int: Rebuilding characteristic time :math:`\tau_{\textrm{REBUILD}}` (see :ref:`boario-math`)."""

        self.overprod_max = alpha_max
        r"""float: Maximum factor of overproduction :math:`\alpha^{\textrm{max}}` (default should be 1.25)."""

        self.overprod_tau = self.n_temporal_units_by_step / alpha_tau
        r"""float: Characteristic time of overproduction :math:`\tau_{\alpha}` in ``n_temporal_units_by_step`` (default should be 365 days)."""

        self.overprod_base = alpha_base
        r"""float: Base value of overproduction factor :math:`\alpha^{b}` (Default to 1.0)."""

        if inventory_dict is None:
            infinite_inventories_sect = (
                [] if infinite_inventories_sect is None else infinite_inventories_sect
            )
            self.inventories = [
                np.inf if sector in infinite_inventories_sect else main_inv_dur
                for sector in self.sectors
            ]
            r"""numpy.ndarray of int: Array :math:`\mathbf{s}` of size :math:`n` (sectors), setting for each input the initial number of ``n_temporal_units_by_step`` of stock for the input. (see :ref:`boario-math`)."""

        else:
            infinite_inventories_sect = (
                [] if infinite_inventories_sect is None else infinite_inventories_sect
            )
            self.inventories = [
                np.inf
                if inventory_dict[k] in ["inf", "Inf", "Infinity", "infinity"]
                or k in infinite_inventories_sect
                else inventory_dict[k]
                for k in sorted(inventory_dict.keys())
            ]

        logger.debug(f"inventories: {self.inventories}")
        logger.debug(f"n_temporal_units_by_step: {self.n_temporal_units_by_step}")
        self.inv_duration = np.array(self.inventories) / self.n_temporal_units_by_step

        if (self.inv_duration <= 1).any():
            logger.warning(
                f"At least one product has inventory duration lower than the numbers of temporal units in one step ({self.n_temporal_units_by_step}), model will set it to 2 by default, but you should probably check this !"
            )
            self.inv_duration[self.inv_duration <= 1] = 2
        #################################################################

        ######## INITIAL MRIO STATE (in step temporality) ###############
        if pym_mrio.Z is None:
            raise ValueError(
                "Z attribute of given MRIO doesn't exist, this is a problem"
            )
        if pym_mrio.Y is None:
            raise ValueError(
                "Y attribute of given MRIO doesn't exist, this is a problem"
            )
        if pym_mrio.x is None:
            raise ValueError(
                "x attribute of given MRIO doesn't exist, this is a problem"
            )
        self._matrix_id = np.eye(self.n_sectors)
        self._matrix_I_sum = np.tile(self._matrix_id, self.n_regions)
        self.Z_0 = pym_mrio.Z.to_numpy()
        r"""numpy.ndarray of float: 2-dim square matrix array :math:`\ioz` of size :math:`(n \times m, n \times m)` representing the daily intermediate (transaction) matrix (see :ref:`boario-math-init`)."""

        self.Z_C = self._matrix_I_sum @ self.Z_0
        r"""numpy.ndarray of float: 2-dim matrix array :math:`\ioz^{\sectorsset}` of size :math:`(n, n \times m)` representing the intermediate (transaction) matrix aggregated by inputs (see :ref:`here <boario-math-z-agg>`)."""

        with np.errstate(divide="ignore", invalid="ignore"):
            self.Z_distrib = np.divide(
                self.Z_0, (np.tile(self.Z_C, (self.n_regions, 1)))
            )
            r"""numpy.ndarray of float: math:`\ioz` normalised by :math:`\ioz^{\sectorsset}`, i.e. representing for each input the share of the total ordered transiting from an industry to another (see :ref:`here <boario-math-z-agg>`)."""

        self.Z_distrib = np.nan_to_num(self.Z_distrib)
        self.Z_0 = pym_mrio.Z.to_numpy() * self.steply_factor

        self.Y_0 = pym_mrio.Y.to_numpy()
        self.Y_C = self._matrix_I_sum @ self.Y_0
        r"""numpy.ndarray of float: 2-dim matrix array :math:`\ioy^{\sectorsset}` of size :math:`(n, m \times \text{number of final demand categories})` representing the final demand matrix aggregated by inputs (see :ref:`here <boario-math-z-agg>`)."""

        with np.errstate(divide="ignore", invalid="ignore"):
            self.Y_distrib = np.divide(
                self.Y_0, (np.tile(self.Y_C, (self.n_regions, 1)))
            )
            r"""numpy.ndarray of float: math:`\ioy` normalised by :math:`\ioy^{\sectorsset}`, i.e. representing for each input the share of the total ordered transiting from an industry to final demand (see :ref:`here <boario-math-z-agg>`)."""

        self.Y_distrib = np.nan_to_num(self.Y_distrib)

        self.Y_0 = pym_mrio.Y.to_numpy() * self.steply_factor
        r"""numpy.ndarray of float: 2-dim array :math:`\ioy` of size :math:`(n \times m, m \times \text{number of final demand categories})` representing the daily final demand matrix."""

        tmp = np.array(pym_mrio.x.T)
        self.X_0 = tmp.flatten() * self.steply_factor
        r"""numpy.ndarray of float: Array :math:`\iox(0)` of size :math:`n \times m` representing the initial daily gross production."""

        del tmp
        value_added = pym_mrio.x.T - pym_mrio.Z.sum(axis=0)
        value_added = value_added.reindex(sorted(value_added.index), axis=0)
        value_added = value_added.reindex(sorted(value_added.columns), axis=1)
        if (value_added < 0).any().any():
            logger.warning(
                f"""Found negative values in the value added, will set to 0. Note that sectors with null value added are not impacted by capital destruction.
                industries with negative VA: {(value_added[value_added < 0].dropna(axis=1)).columns}
                """
            )

        value_added[value_added < 0] = 0.0
        self.gdp_df = value_added.groupby("region", axis=1).sum()
        r"""pandas.DataFrame: Dataframe of the total GDP of each region of the model"""

        self.VA_0 = value_added.to_numpy().flatten()
        r"""numpy.ndarray of float: Array :math:`\iov` of size :math:`n \times m` representing the total value added for each sectors."""

        self.tech_mat = (self._matrix_I_sum @ pym_mrio.A).to_numpy()  # type: ignore #to_numpy is not superfluous !
        r"""numpy.ndarray of float: 2-dim array :math:`\ioa` of size :math:`(n \times m, n \times m)` representing the technical coefficients matrix."""

        if productive_capital_vector is not None:
            self.k_stock = productive_capital_vector
            """numpy.ndarray of float: Array of size :math:`n \times m` representing the estimated stock of capital of each industry."""

            if isinstance(self.k_stock, pd.DataFrame):
                self.k_stock = self.k_stock.squeeze().sort_index().to_numpy()
        elif productive_capital_to_VA_dict is None:
            logger.warning(f"No capital to VA dictionary given, considering 4/1 ratio")
            self.kstock_ratio_to_VA = 4
            self.k_stock = self.VA_0 * self.kstock_ratio_to_VA
        else:
            kratio = productive_capital_to_VA_dict
            kratio_ordered = [kratio[k] for k in sorted(kratio.keys())]
            self.kstock_ratio_to_VA = np.tile(np.array(kratio_ordered), self.n_regions)
            self.k_stock = self.VA_0 * self.kstock_ratio_to_VA
        if value_added.ndim > 1:
            self.gdp_share_sector = (
                self.VA_0
                / value_added.sum(axis=0).groupby("region").transform("sum").to_numpy()
            )
        else:
            self.gdp_share_sector = (
                self.VA_0 / value_added.groupby("region").transform("sum").to_numpy()
            )
        self.gdp_share_sector = self.gdp_share_sector.flatten()
        r"""numpy.ndarray of float: Array of size :math:`n \times m` representing the estimated share of a sector in its regional economy."""

        self.matrix_share_thresh = (
            self.Z_C > np.tile(self.X_0, (self.n_sectors, 1)) * TECHNOLOGY_THRESHOLD
        )  # [n_sectors, n_regions*n_sectors]
        r"""numpy.ndarray of float: 2-dim square matrix array of size :math:`(n , n \times m)` representing the threshold under which an input is not considered being an input (0.00001)."""
        #################################################################

        ####### SIMULATION VARIABLES ####################################
        self.overprod = np.full(
            (self.n_regions * self.n_sectors), self.overprod_base, dtype=np.float64
        )
        r"""numpy.ndarray of float: Array of size :math:`n \times m` representing the overproduction coefficients vector :math:`\mathbf{\alpha}(t)`."""

        with np.errstate(divide="ignore", invalid="ignore"):
            self.matrix_stock = (
                np.tile(self.X_0, (self.n_sectors, 1)) * self.tech_mat
            ) * self.inv_duration[:, np.newaxis]
        self.matrix_stock = np.nan_to_num(self.matrix_stock, nan=np.inf, posinf=np.inf)
        r"""numpy.ndarray of float: 2-dim square matrix array :math:`\ioinv` of size :math:`(n \times m, n \times m)` representing the stock of inputs (see :ref:`boario-math-init`)."""

        self.matrix_stock_0 = self.matrix_stock.copy()
        self.matrix_orders = self.Z_0.copy()
        r"""numpy.ndarray of float: 2-dim square matrix array :math:`\ioorders` of size :math:`(n \times m, n \times m)` representing the matrix of orders."""

        self.production = self.X_0.copy()
        r"""numpy.ndarray of float: Array :math:`\iox(t)` of size :math:`n \times m` representing the current gross production."""

        self.intmd_demand = self.Z_0.copy()
        self.final_demand = self.Y_0.copy()
        self.rebuilding_demand = None
        self.productive_capital_lost = np.zeros(self.production.shape)
        r"""numpy.ndarray of float: Array of size :math:`n \times m` representing the estimated stock of capital currently destroyed for each industry."""

        self.order_type = order_type
        #################################################################

        ################## SIMULATION TRACKING VARIABLES ################
        self.in_shortage = False
        r"""Boolean stating if at least one industry is in shortage (i.e.) if at least one of its inputs inventory is low enough to reduce production."""

        self.had_shortage = False
        r"""Boolean stating if at least one industry was in shortage at some point."""

        self.rebuild_prod = np.zeros(shape=self.X_0.shape)
        r"""numpy.ndarray of float: Array of size :math:`n \times m` representing the remaining stock of rebuilding demand asked of each industry."""

        self.final_demand_not_met = np.zeros(self.Y_0.shape)
        r"""numpy.ndarray of float: Array of size :math:`n \times m` representing the final demand that could not be met at current step for each industry."""

        #################################################################

        ### Internals
        self._prod_delta_type = None

        #### POST INIT ####
        ### Event Class Attribute setting
        logger.debug(
            f"Setting possible regions (currently: {Event.possible_regions}) to: {self.regions}"
        )
        Event.possible_regions = pd.CategoricalIndex(
            self.regions, name="region", copy=True
        )
        logger.debug(f"Possible regions is now {Event.possible_regions}")
        Event.regions_idx = np.arange(self.n_regions)
        Event.possible_sectors = pd.CategoricalIndex(
            self.sectors, name="sector", copy=True
        )
        Event.possible_final_demand_cat = pd.CategoricalIndex(
            pym_mrio.get_Y_categories(), name="final demand", copy=True
        )
        Event.sectors_idx = np.arange(self.n_sectors)
        Event.z_shape = self.Z_0.shape
        Event.y_shape = self.Y_0.shape
        Event.x_shape = self.X_0.shape
        Event.model_monetary_factor = monetary_factor
        Event.sectors_gva_shares = self.gdp_share_sector.copy()
        Event.Z_distrib = self.Z_distrib.copy()
        Event.Y_distrib = self.Y_distrib.copy()
        # logger.warning(f"{value_added}")
        Event.gva_df = value_added.iloc[0].copy()

        meta = pym_mrio.meta.metadata
        assert isinstance(meta, dict)
        # meta = {"name":"unnamed", "description":"", "system":"unknown","version":"unknown"}
        try:
            Event.mrio_name = "_".join(
                [
                    str(meta["name"]),
                    str(meta["description"]),
                    str(meta["system"]),
                    str(meta["version"]),
                ]
            )
        except TypeError:
            Event.mrio_name = "custom - method WIP"

        # initialize those (it's not very nice, but otherwise python complains)
        self._indus_rebuild_demand_tot: npt.NDArray = np.array([])
        self._house_rebuild_demand_tot: npt.NDArray = np.array([])
        self._indus_rebuild_demand: npt.NDArray = np.array([])
        self._house_rebuild_demand: npt.NDArray = np.array([])
        self._tot_rebuild_demand: npt.NDArray = np.array([])
        self._productive_capital_lost: npt.NDArray = np.zeros(self.VA_0.shape)
        self._prod_cap_delta_productive_capital: npt.NDArray = np.zeros(
            len(self.industries)
        )
        self._prod_cap_delta_arbitrary: npt.NDArray = np.zeros(len(self.industries))
        self._prod_cap_delta_tot: npt.NDArray = np.zeros(len(self.industries))

    ## Properties

    @property
    def tot_rebuild_demand(self) -> npt.NDArray:
        r"""Returns current total rebuilding demand (as the sum of rebuilding demand addressed to each industry)"""
        tmp = []
        logger.debug("Trying to return tot_rebuilding demand")
        if np.any(self._indus_rebuild_demand_tot > 0):
            tmp.append(self._indus_rebuild_demand_tot)
        if np.any(self._house_rebuild_demand_tot > 0):
            tmp.append(self._house_rebuild_demand_tot)
        if tmp:
            ret = np.concatenate(tmp, axis=1).sum(axis=1)
            self._tot_rebuild_demand = ret
        else:
            self._tot_rebuild_demand = np.array([])
        return self._tot_rebuild_demand

    @tot_rebuild_demand.setter
    def tot_rebuild_demand(self, source: list[EventKapitalRebuild]):
        r"""Computes and updates total rebuilding demand based on a list of events.

        Compute and update rebuilding demand for the given list of events. Only events
        tagged as rebuildable are accounted for. Both `house_rebuild_demand` and
        `indus_rebuild_demand` are updated.

        Parameters
        ----------
        events : 'list[EventKapitalRebuild]'
            A list of EventKapitalRebuild objects
        """
        logger.debug(f"Trying to set tot_rebuilding demand from {source}")
        if not isinstance(source, list):
            ValueError(
                f"Setting tot_rebuild_demand can only be done with a list of events, not a {type(source)}"
            )
        self.house_rebuild_demand = source
        self.indus_rebuild_demand = source

    @property
    def house_rebuild_demand(self) -> npt.NDArray:
        r"""Returns household rebuilding demand matrix.

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\ioy`, containing the sum of all currently
            rebuildable final demand stock.
        """

        return self._house_rebuild_demand

    @property
    def house_rebuild_demand_tot(self) -> npt.NDArray:
        r"""Returns total household rebuilding demand vector.

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\iox`, containing the sum of all currently
            rebuildable households demands.
        """

        return self._house_rebuild_demand_tot

    @house_rebuild_demand.setter
    def house_rebuild_demand(self, source: list[EventKapitalRebuild]):
        r"""Computes rebuild demand from household from a list of events

        Computes and sets rebuilding final households demand for the given list of events
        by summing the `rebuilding_demand_house member` of each event and multiplying it
        by the characteristic rebuilding time (inverse of its value). Only events
        tagged as rebuildable are accounted for.

        If the event has its own rebuilding characteristic time :math:`\tau_{\textrm{REBUILD}}` it is applied,
        else the model rebuilding characteristic time is used.

        Parameters
        ----------
        events : 'list[Event]'
            A list of Event objects

        """
        tmp = []
        for evnt in source:
            if evnt.rebuildable:
                if evnt.rebuild_tau is None:
                    rebuild_tau = self.rebuild_tau
                else:
                    rebuild_tau = evnt.rebuild_tau
                    warn_once(logger, "Event has a custom rebuild_tau")
                if len(evnt.rebuilding_demand_house) > 0:
                    tmp.append(
                        evnt.rebuilding_demand_house
                        * (self.n_temporal_units_by_step / rebuild_tau)
                    )
        if not tmp:
            self._house_rebuild_demand = np.array([])
            self._house_rebuild_demand_tot = np.array([])
        else:
            house_reb_dem = np.stack(tmp, axis=-1)
            tot = house_reb_dem.sum(axis=1)
            self._house_rebuild_demand = house_reb_dem
            self._house_rebuild_demand_tot = tot

    @property
    def indus_rebuild_demand(self) -> npt.NDArray:
        r"""Returns industrial rebuilding demand matrix.

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\ioz`, containing the sum of all currently
            rebuildable intermediate demand stock.
        """

        return self._indus_rebuild_demand

    @property
    def indus_rebuild_demand_tot(self) -> npt.NDArray:
        r"""Returns total industrial rebuilding demand vector.

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\iox`, containing the sum of all currently
            rebuildable intermediate demands.
        """

        return self._indus_rebuild_demand_tot

    @indus_rebuild_demand.setter
    def indus_rebuild_demand(self, source: list[EventKapitalRebuild]):
        r"""Computes rebuild demand from economic sectors (i.e. not households)

        Computes and sets rebuilding 'intermediate demand' for the given list of events
        by summing the `rebuildind_demand_indus` of each event. Only events
        tagged as rebuildable are accounted for.

        Parameters
        ----------
        events : 'list[Event]'
            A list of Event objects

        """

        tmp = []
        for evnt in source:
            if evnt.rebuildable:
                if evnt.rebuild_tau is None:
                    rebuild_tau = self.rebuild_tau
                else:
                    rebuild_tau = evnt.rebuild_tau
                    warn_once(logger, "Event has a custom rebuild_tau")
                tmp.append(
                    evnt.rebuilding_demand_indus
                    * (self.n_temporal_units_by_step / rebuild_tau)
                )

        if not tmp:
            self._indus_rebuild_demand = np.array([])
            self._indus_rebuild_demand_tot = np.array([])
        else:
            indus_reb_dem = np.stack(tmp, axis=-1)
            self._indus_rebuild_demand = indus_reb_dem
            self._indus_rebuild_demand_tot = indus_reb_dem.sum(axis=1)
            # logger.debug(f"Setting indus_rebuild_demand_tot to {indus_reb_dem}")

    @property
    def productive_capital_lost(self) -> npt.NDArray:
        r"""Returns current stock of destroyed capital

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\iox`, containing the "stock"
        of capital currently destroyed for each industry.
        """

        return self._productive_capital_lost

    @productive_capital_lost.setter
    def productive_capital_lost(
        self, source: list[EventKapitalDestroyed] | npt.NDArray
    ) -> None:
        r"""Computes current capital lost and update production delta accordingly.

        Computes and sets the current stock of capital lost by each industry of
        the model due to the given list of events. Also update the production
        capacity lost accordingly by computing the ratio of capital lost of
        capital stock.

        Parameters
        ----------
        source : list[EventKapitalDestroyed] | npt.NDArray
            Either a list of events to consider for the destruction
        of capital or directly a vector of destroyed capital for each industry.

        """

        logger.debug("Updating productive_capital lost from list of events")
        if isinstance(source, list):
            if source:
                self._productive_capital_lost = np.add.reduce(
                    np.array(
                        [
                            e.regional_sectoral_productive_capital_destroyed
                            for e in source
                        ]
                    )
                )
            else:
                self._productive_capital_lost = np.zeros(self.VA_0.shape)
        elif isinstance(source, np.ndarray):
            self._productive_capital_lost = source

        productivity_loss_from_capital_destroyed = np.zeros(
            shape=self._productive_capital_lost.shape
        )
        np.divide(
            self._productive_capital_lost,
            self.k_stock,
            out=productivity_loss_from_capital_destroyed,
            where=self.k_stock != 0,
        )
        logger.debug("Updating production delta from productive_capital loss")
        self._prod_cap_delta_productive_capital = (
            productivity_loss_from_capital_destroyed
        )
        if (self._prod_cap_delta_productive_capital > 0.0).any():
            if self._prod_delta_type is None:
                self._prod_delta_type = "from_productive_capital"
            elif self._prod_delta_type == "from_arbitrary":
                self._prod_delta_type = "mixed_from_productive_capital_from_arbitrary"

    @property
    def prod_cap_delta_arbitrary(self) -> npt.NDArray:
        r"""Return the possible "arbitrary" production capacity lost vector if
        it was set.

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\iox`, stating the amount of production
        capacity lost arbitrarily (ie exogenous).
        """
        return self._prod_cap_delta_arbitrary

    @prod_cap_delta_arbitrary.setter
    def prod_cap_delta_arbitrary(self, source: list[EventArbitraryProd] | npt.NDArray):
        """Computes and sets the loss of production capacity from "arbitrary" sources.

        Parameters
        ----------
        source : list[Event] | npt.NDArray
            Either a list of Event objects with arbitrary production losses
            set, or directly a vector of production capacity loss.

        """

        if isinstance(source, list):
            event_arb = np.array(
                [
                    ev.prod_cap_delta_arbitrary
                    for ev in source
                    if len(ev.prod_cap_delta_arbitrary) > 0
                ]
            )
            if event_arb.size == 0:
                self._prod_cap_delta_arbitrary = np.zeros(shape=self.X_0.shape)
            else:
                self._prod_cap_delta_arbitrary = np.maximum.reduce(event_arb)
        else:
            self._prod_capt_delta_arbitrary = source
        assert self._prod_cap_delta_arbitrary is not None
        if (self._prod_cap_delta_arbitrary > 0.0).any():
            if self._prod_delta_type is None:
                self._prod_delta_type = "from_arbitrary"
            elif self._prod_delta_type == "from_productive_capital":
                self._prod_delta_type = "mixed_from_productive_capital_from_arbitrary"

    @property
    def prod_cap_delta_productive_capital(self) -> npt.NDArray:
        r"""Return the possible production capacity lost due to capital destroyed vector if
        it was set.

        Returns
        -------
        npt.NDArray
            An array of same shape as math:`\iox`, stating the amount of production
        capacity lost due to capital destroyed.
        """

        return self._prod_cap_delta_productive_capital

    @property
    def prod_cap_delta_tot(self) -> npt.NDArray:
        r"""Computes and return total current production delta.

        Returns
        -------
        npt.NDArray
            The total production delta (ie share of production capacity lost)
            for each industry.

        """

        logger.debug("Trying to retrieve current production delta")
        tmp = []
        if self._prod_delta_type is None:
            raise AttributeError("Production delta doesn't appear to be set yet.")
        elif self._prod_delta_type == "from_productive_capital":
            logger.debug(
                "Production delta is only set from productive_capital destruction"
            )
            tmp.append(self._prod_cap_delta_productive_capital)
        elif self._prod_delta_type == "from_arbitrary":
            logger.debug("Production delta is only set from arbitrary delta")
            tmp.append(self._prod_cap_delta_arbitrary)
        elif self._prod_delta_type == "mixed_from_productive_capital_from_arbitrary":
            logger.debug(
                "Production delta is a mixed form of productive_capital destruction and arbitrary delta"
            )
            tmp.append(self._prod_cap_delta_productive_capital)
            tmp.append(self._prod_cap_delta_arbitrary)
        else:
            raise NotImplementedError(
                f"Production delta type {self._prod_delta_type} not recognised"
            )
        # tmp.append(np.ones(shape=self.X_0.shape))
        # logger.debug("tmp: {}".format(tmp))
        self._prod_cap_delta_tot = np.amax(np.stack(tmp, axis=-1), axis=1)
        assert (
            self._prod_cap_delta_tot.shape == self.X_0.shape
        ), f"expected shape {self.X_0.shape}, received {self._prod_cap_delta_tot.shape}"
        return self._prod_cap_delta_tot

    @prod_cap_delta_tot.setter
    def prod_cap_delta_tot(self, source: list[Event]):
        r"""Computes and sets the loss of production capacity from both "arbitrary" sources and
        capital destroyed.

        Parameters
        ----------
        source : list[Event]
            A list of Event objects.

        """

        logger.debug("Updating total production delta from list of events")
        if not isinstance(source, list):
            ValueError(
                f"Setting prod_cap_delta_tot can only be done with a list of events, not a {type(source)}"
            )

        self.productive_capital_lost = [
            event for event in source if isinstance(event, EventKapitalDestroyed)
        ]
        self.prod_cap_delta_arbitrary = [
            event for event in source if isinstance(event, EventArbitraryProd)
        ]

    def update_system_from_events(self, events: list[Event]) -> None:
        """Update model variables according to given list of events

        Computes and sets both the total production delta from all events, and the total rebuilding demand.

        Parameters
        ----------
        events : 'list[Event]'
            List of events (as Event objects) to consider.

        """
        logger.debug("Updating system from list of events")
        self.prod_cap_delta_tot = events
        self.tot_rebuild_demand = [
            event for event in events if isinstance(event, EventKapitalRebuild)
        ]

    @property
    def production_cap(self) -> npt.NDArray:
        r"""Compute and update production capacity.

        Compute and update production capacity from current total production delta and overproduction.

        .. math::

            x^{Cap}_{f}(t) = \alpha_{f}(t) (1 - \Delta_{f}(t)) x_{f}(t)

        Raises
        ------
        ValueError
            Raised if any industry has negative production.
        """
        production_cap = self.X_0.copy()
        if self._prod_delta_type is not None:
            prod_delta_tot = self.prod_cap_delta_tot.copy()
            if (prod_delta_tot > 0.0).any():
                production_cap = production_cap * (1 - prod_delta_tot)
        if (self.overprod > 1.0).any():
            production_cap = production_cap * self.overprod
        if (production_cap < 0).any():
            raise ValueError(
                "Production capacity was found negative for at least on industry"
            )
        return production_cap

    @property
    def total_demand(self) -> npt.NDArray:
        r"""Computes and returns total demand as the sum of intermediate demand (orders), final demand, and possible rebuilding demand."""

        if (self.matrix_orders < 0).any():
            raise RuntimeError("Some matrix orders are negative which shouldn't happen")

        tot_dem = self.matrix_orders.sum(axis=1) + self.final_demand.sum(axis=1)
        if (tot_dem < 0).any():
            raise RuntimeError("Some total demand are negative which shouldn't happen")
        if len(self.tot_rebuild_demand) > 0:
            tot_dem += self.tot_rebuild_demand
        return tot_dem

    @property
    def production_opt(self) -> npt.NDArray:
        r"""Computes and returns "optimal production" :math:`\iox^{textrm{Opt}}`, as the per industry minimum between
        total demand and production capacity.

        """

        return np.fmin(self.total_demand, self.production_cap)

    @property
    def inventory_constraints_opt(self) -> npt.NDArray:
        r"""Computes and returns inventory constraints for "optimal production" (see :meth:`calc_inventory_constraints`)"""

        return self.calc_inventory_constraints(self.production_opt)

    @property
    def inventory_constraints_act(self) -> npt.NDArray:
        r"""Computes and returns inventory constraints for "actual production" (see :meth:`calc_inventory_constraints`)"""
        return self.calc_inventory_constraints(self.production)

    def calc_production(self, current_temporal_unit: int) -> npt.NDArray:
        r"""Computes and updates actual production. See :ref:`boario-math-prod`.

        1. Computes ``production_opt`` and ``inventory_constraints`` as :

        .. math::
           :nowrap:

                \begin{alignat*}{4}
                      \iox^{\textrm{Opt}}(t) &= (x^{\textrm{Opt}}_{f}(t))_{f \in \firmsset} &&= \left ( \min \left ( d^{\textrm{Tot}}_{f}(t), x^{\textrm{Cap}}_{f}(t) \right ) \right )_{f \in \firmsset} && \text{Optimal production}\\
                      \mathbf{\ioinv}^{\textrm{Cons}}(t) &= (\omega^{\textrm{Cons},f}_p(t))_{\substack{p \in \sectorsset\\f \in \firmsset}} &&=
                    \begin{bmatrix}
                        \tau^{1}_1 & \hdots & \tau^{p}_1 \\
                        \vdots & \ddots & \vdots\\
                        \tau^1_n & \hdots & \tau^{p}_n
                    \end{bmatrix}
                    \odot \begin{bmatrix} \iox^{\textrm{Opt}}(t)\\ \vdots\\ \iox^{\textrm{Opt}}(t) \end{bmatrix} \odot \ioa^{\sectorsset} && \text{Inventory constraints} \\
                    &&&= \begin{bmatrix}
                        \tau^{1}_1 x^{\textrm{Opt}}_{1}(t) a_{11} & \hdots & \tau^{p}_1 x^{\textrm{Opt}}_{p}(t) a_{1p}\\
                        \vdots & \ddots & \vdots\\
                        \tau^1_n x^{\textrm{Opt}}_{1}(t) a_{n1} & \hdots & \tau^{p}_n x^{\textrm{Opt}}_{p}(t) a_{np}
                    \end{bmatrix} && \\
                \end{alignat*}

        2. If stocks do not meet ``inventory_constraints`` for any inputs, then decrease production accordingly :

        .. math::
           :nowrap:

                \begin{alignat*}{4}
                    \iox^{a}(t) &= (x^{a}_{f}(t))_{f \in \firmsset} &&= \left \{ \begin{aligned}
                                                           & x^{\textrm{Opt}}_{f}(t) & \text{if $\omega_{p}^f(t) \geq \omega^{\textrm{Cons},f}_p(t)$}\\
                                                           & x^{\textrm{Opt}}_{f}(t) \cdot \min_{p \in \sectorsset} \left ( \frac{\omega^s_{p}(t)}{\omega^{\textrm{Cons,f}}_p(t)} \right ) & \text{if $\omega_{p}^f(t) < \omega^{\textrm{Cons},f}_p(t)$}
                                                           \end{aligned} \right. \quad &&
                \end{alignat*}

        Also warns in logs if such shortages happen.


        Parameters
        ----------
        current_temporal_unit : int
            current step number

        """
        # 1.
        production_opt = self.production_opt.copy()
        inventory_constraints = self.inventory_constraints_opt.copy()
        # 2.
        if (
            stock_constraint := (self.matrix_stock < inventory_constraints)
            * self.matrix_share_thresh
        ).any():
            if not self.in_shortage:
                logger.info(
                    f"At least one industry entered shortage regime. (step:{current_temporal_unit})"
                )
            self.in_shortage = True
            self.had_shortage = True
            production_ratio_stock = np.ones(shape=self.matrix_stock.shape)
            np.divide(
                self.matrix_stock,
                inventory_constraints,
                out=production_ratio_stock,
                where=(self.matrix_share_thresh * (inventory_constraints != 0)),
            )
            production_ratio_stock[production_ratio_stock > 1] = 1
            if (production_ratio_stock < 1).any():
                production_max = (
                    np.tile(production_opt, (self.n_sectors, 1))
                    * production_ratio_stock
                )
                assert not (np.min(production_max, axis=0) < 0).any()
                self.production = np.min(production_max, axis=0)
            else:
                assert not (production_opt < 0).any()
                self.production = production_opt
        else:
            if self.in_shortage:
                self.in_shortage = False
                logger.info(
                    f"All industries exited shortage regime. (step:{current_temporal_unit})"
                )
            assert not (production_opt < 0).any()
            self.production = production_opt
        return stock_constraint

    def calc_inventory_constraints(self, production: npt.NDArray) -> npt.NDArray:
        r"""Compute inventory constraints (no psi parameter, for the psi version,
        the recommended one, see :meth:`~boario.extended_models.ARIOPsiModel.calc_inventory_constraints`)

        See :meth:`calc_production` for how inventory constraints are computed.

        Parameters
        ----------
        production : npt.NDArray
            The production vector to consider.

        Returns
        -------
        npt.NDArray
            For each input, for each industry, the size of the inventory required to produce at `production` level
        for the duration goal (`inv_duration`).

        """
        inventory_constraints = np.tile(production, (self.n_sectors, 1)) * self.tech_mat
        tmp = np.tile(
            np.nan_to_num(self.inv_duration, posinf=0.0)[:, np.newaxis],
            (1, self.n_regions * self.n_sectors),
        )
        return inventory_constraints * tmp

    def distribute_production(
        self,
        rebuildable_events: list[EventKapitalRebuild],
        scheme: str = "proportional",
    ) -> list[EventKapitalRebuild]:
        r"""Production distribution module

    #. Computes rebuilding demand for each rebuildable events (applying the `rebuild_tau` characteristic time)

    #. Creates/Computes total demand matrix (Intermediate + Final + Rebuild)

    #. Assesses if total demand is greater than realized production, hence requiring rationing

    #. Distributes production proportionally to demand such that :

        .. math::
           :nowrap:

               \begin{alignat*}{4}
                   &\ioorders^{\textrm{Received}}(t) &&= \left (\frac{o_{ff'}(t)}{d^{\textrm{Tot}}_f(t)} \cdot x^a_f(t) \right )_{f,f'\in \firmsset}\\
                   &\ioy^{\textrm{Received}}(t) &&= \left ( \frac{y_{f,c}}{d^{\textrm{Tot}}_f(t)}\cdot x^a_f(t) \right )_{f\in \firmsset, c \in \catfdset}\\
                   &\Damage^{\textrm{Repaired}}(t) &&= \left ( \frac{\gamma_{f,c}}{d^{\textrm{Tot}}_f(t)} \cdot x^a_f(t) \right )_{f\in \firmsset, c \in \catfdset}\\
               \end{alignat*}

        Where :

        - :math:`\ioorders^{\textrm{Received}}(t)` is the received orders matrix,

        - :math:`\ioy^{\textrm{Received}}(t)` is the final demand received matrix,

        - :math:`\Damage^{\textrm{Repared}}(t)` is the rebuilding/repair achieved matrix,

        - :math:`d^{\textrm{Tot}}_f(t)` is the total demand to industry :math:`f`,

        - :math:`x^a_f(t)` is :math:`f`'s realized production,

        - :math:`o_{ff'}(t)` is the quantity of product ordered by industry :math:`f'` to industry :math:`f`,

        - :math:`y_{fc}(t)` is the quantity of product ordered by household :math:`c` to industry :math:`f`,

        - :math:`\gamma_{fc}(t)` is the repaired/rebuilding demand ordered to :math:`f`.

    #. Updates stocks matrix. (Only if `np.allclose(stock_add, stock_use).all()` is false)

        .. math::
           :nowrap:

               \begin{alignat*}{4}
                   &\ioinv(t+1) &&= \ioinv(t) + \left ( \mathbf{I}_{\textrm{sum}} \cdot \ioorders^{\textrm{Received}}(t) \right ) - \left ( \colvec{\iox^{\textrm{a}}(t)}{\iox^{\textrm{a}}(t)} \odot \ioa^{\sectorsset} \right )\\
               \end{alignat*}

        Where :

        - :math:`\ioinv` is the inventory matrix,
        - :math:`\mathbf{I}_{\textrm{sum}}` is a row summation matrix,
        - :math:`\ioa^{\sectorsset}` is the (input not specific to region) technical coefficients matrix.

    #. Computes final demand not met due to rationing and write it.

    #. Updates rebuilding demand for each event (by substracting distributed production)

    Parameters
    ----------
    current_temporal_unit : int
        Current temporal unit (day|week|... depending on parameters) (required to write the final demand not met)
    events : 'list[Event]'
        List of rebuildable events
    scheme : str
        Placeholder for future distribution scheme
    separate_rebuilding : bool
        Currently unused.

    Returns
    -------
    list[Event]
        The list of events to remove from current events (as they are totally rebuilt)

    Raises
    ------
    RuntimeError
        If negative values are found in places there's should not be any
    ValueError
        If an attempt to run an unimplemented distribution scheme is tried

"""

        if scheme != "proportional":
            raise ValueError(f"Scheme {scheme} is currently not implemented")

        # list_of_demands = [self.matrix_orders, self.final_demand]
        ## 1. Calc demand from rebuilding requirements (with characteristic time rebuild_tau)
        house_reb_dem_per_event = (
            house_reb_dem_tot_per_event
        ) = indus_reb_dem_per_event = indus_reb_dem_tot_per_event = None

        if rebuildable_events:
            logger.debug("There are rebuildable events")
            n_events = len(rebuildable_events)
            tot_rebuilding_demand_summed = self.tot_rebuild_demand.copy()
            # debugging assert
            if not tot_rebuilding_demand_summed.shape == self.X_0.shape:
                raise RuntimeError(
                    f"received {tot_rebuilding_demand_summed.shape}, expected {self.X_0.shape}"
                )
            indus_reb_dem_tot_per_event = self.indus_rebuild_demand_tot.copy()
            indus_reb_dem_per_event = self.indus_rebuild_demand.copy()

            # expected shape assert (debug also)
            exp_shape_indus_per_event = (
                self.n_sectors * self.n_regions,
                self.n_sectors * self.n_regions,
                n_events,
            )
            exp_shape_indus_tot_per_event = (self.n_sectors * self.n_regions, n_events)
            assert (
                indus_reb_dem_per_event.shape == exp_shape_indus_per_event
            ), "expected shape is {}, given shape is {}".format(
                exp_shape_indus_per_event, indus_reb_dem_per_event.shape
            )
            assert (
                indus_reb_dem_tot_per_event.shape == exp_shape_indus_tot_per_event
            ), "expected shape is {}, given shape is {}".format(
                exp_shape_indus_tot_per_event, indus_reb_dem_tot_per_event.shape
            )

            house_reb_dem_tot_per_event = self.house_rebuild_demand_tot.copy()
            house_reb_dem_per_event = self.house_rebuild_demand.copy()

            # expected shape assert (debug also)
            exp_shape_house = (
                self.n_sectors * self.n_regions,
                self.n_fd_cat * self.n_regions,
                n_events,
            )
            exp_shape_house_tot = (self.n_sectors * self.n_regions, n_events)
            assert house_reb_dem_per_event.shape == exp_shape_house
            assert house_reb_dem_tot_per_event.shape == exp_shape_house_tot
            h_reb = (house_reb_dem_tot_per_event > 0).any()
            ind_reb = (indus_reb_dem_tot_per_event > 0).any()
        else:
            tot_rebuilding_demand_summed = np.zeros(shape=self.X_0.shape)
            h_reb = False
            ind_reb = False

        ## 2. Concat to have total demand matrix (Intermediate + Final + Rebuild)
        tot_demand = np.concatenate(
            [
                self.matrix_orders,
                self.final_demand,
                np.expand_dims(tot_rebuilding_demand_summed, 1),
            ],
            axis=1,
        )
        ## 3. Does production meet total demand
        logger.debug(f"tot demand shape : {tot_demand.shape}")
        rationing_required = (self.production - tot_demand.sum(axis=1)) < (
            -1 / self.monetary_factor
        )
        rationning_mask = np.tile(
            rationing_required[:, np.newaxis], (1, tot_demand.shape[1])
        )
        demand_share = np.full(tot_demand.shape, 0.0)
        tot_dem_summed = np.expand_dims(
            np.sum(tot_demand, axis=1, where=rationning_mask), 1
        )
        # Get demand share
        np.divide(
            tot_demand, tot_dem_summed, where=(tot_dem_summed != 0), out=demand_share
        )
        distributed_production = tot_demand.copy()
        # 4. distribute production proportionally to demand
        np.multiply(
            demand_share,
            np.expand_dims(self.production, 1),
            out=distributed_production,
            where=rationning_mask,
        )
        intmd_distribution = distributed_production[
            :, : self.n_sectors * self.n_regions
        ]
        # Stock use is simply production times technical coefs
        stock_use = np.tile(self.production, (self.n_sectors, 1)) * self.tech_mat
        if (stock_use < 0).any():
            raise RuntimeError(
                "Stock use contains negative values, this should not happen"
            )
        # 5. Restock is the production from each supplier, summed.
        stock_add = self._matrix_I_sum @ intmd_distribution
        if (stock_add < 0).any():
            raise RuntimeError(
                "stock_add (restocking) contains negative values, this should not happen"
            )
        if not np.allclose(stock_add, stock_use):
            self.matrix_stock = self.matrix_stock - stock_use + stock_add
            if (self.matrix_stock < 0).any():
                raise RuntimeError(
                    "matrix_stock contains negative values, this should not happen"
                )

        # 6. Compute final demand not met due to rationing
        final_demand_not_met = (
            self.final_demand
            - distributed_production[
                :,
                self.n_sectors
                * self.n_regions : (
                    self.n_sectors * self.n_regions + self.n_fd_cat * self.n_regions
                ),
            ]
        )
        final_demand_not_met = final_demand_not_met.sum(axis=1)
        # avoid -0.0 (just in case)
        final_demand_not_met[final_demand_not_met == 0.0] = 0.0
        self.final_demand_not_met = final_demand_not_met.copy()

        # 7. Compute production delivered to rebuilding
        logger.debug(f"distributed prod shape : {distributed_production.shape}")
        rebuild_prod = (
            distributed_production[
                :, (self.n_sectors * self.n_regions + self.n_fd_cat * self.n_regions) :
            ]
            .copy()
            .flatten()
        )
        self.rebuild_prod = rebuild_prod.copy()

        logger.debug(
            f"tot_rebuilding_demand_summed: {pd.Series(tot_rebuilding_demand_summed, index=self.industries)}"
        )

        if h_reb and ind_reb:
            logger.debug(
                "There are both household rebuilding demand and industry rebuilding demand"
            )
            logger.debug(
                f"indus_reb_dem_tot_per_event: {pd.DataFrame(indus_reb_dem_tot_per_event, index=self.industries)}"
            )
            logger.debug(
                f"house_reb_dem_tot_per_event: {pd.DataFrame(house_reb_dem_tot_per_event, index=self.industries)}"
            )
            logger.debug(
                f"dem_tot_per_event: {pd.DataFrame(house_reb_dem_tot_per_event+indus_reb_dem_tot_per_event, index=self.industries)}"  # type: ignore
            )

            tot_reb_dem_divider = np.tile(
                tot_rebuilding_demand_summed[:, np.newaxis],
                (1, indus_reb_dem_tot_per_event.shape[1]),
            )
            indus_shares = np.divide(
                indus_reb_dem_tot_per_event,
                tot_reb_dem_divider,
                out=np.zeros_like(indus_reb_dem_tot_per_event),
                where=tot_reb_dem_divider != 0,
            )
            house_shares = np.divide(
                house_reb_dem_tot_per_event,
                tot_reb_dem_divider,
                out=np.zeros_like(house_reb_dem_tot_per_event),
                where=tot_reb_dem_divider != 0,
            )
            # np.around(house_shares, 6, out=house_shares)
            # np.around(indus_shares, 6, out=indus_shares)
            logger.debug(
                f"indus_shares: {pd.DataFrame(indus_shares, index=self.industries)}"
            )
            logger.debug(
                f"house_shares: {pd.DataFrame(house_shares, index=self.industries)}"
            )

            #             tot_reb = pd.Series(tot_rebuilding_demand_summed, index=self.industries)
            #             indus_reb_dem_tot = pd.DataFrame(indus_reb_dem_tot_per_event, index=self.industries)[0]
            #             house_reb_dem_tot = pd.DataFrame(house_reb_dem_tot_per_event, index=self.industries)[0]
            #             np.around(indus_shares,10,indus_shares)
            #             np.around(house_shares,10,house_shares)
            #             np.around(tmp,10,tmp)
            #             indus_df = pd.DataFrame(indus_shares,index=self.industries)[0]
            #             house_df = pd.DataFrame(house_shares,index=self.industries)[0]
            #             tmp_df=pd.DataFrame(tmp, index=self.industries)[0]

            #             debug_df = pd.concat([tot_reb, indus_reb_dem_tot, house_reb_dem_tot, tot_reb-(house_reb_dem_tot+indus_reb_dem_tot), indus_df, house_df, tmp_df], axis=1)
            #             #logger.info(debug_df[debug_df.iloc[:,6]!=0].to_string())
            #             debug_str = f"""
            # {debug_df[debug_df.iloc[:,6]!=0].to_string()}
            # """
            assert np.allclose(
                (indus_shares + house_shares)[tot_rebuilding_demand_summed != 0], 1.0
            )

        elif h_reb:
            house_shares = np.ones(house_reb_dem_tot_per_event.shape)
            indus_shares = np.zeros(indus_reb_dem_tot_per_event.shape)
        elif ind_reb:
            house_shares = np.zeros(house_reb_dem_tot_per_event.shape)
            indus_shares = np.ones(indus_reb_dem_tot_per_event.shape)
            # logger.debug("indus_shares: {}".format(indus_shares.shape))
        else:
            return []

        logger.debug(f"rebuild_prod: {rebuild_prod}")
        logger.debug(f"indus_shares: {indus_shares}")

        indus_rebuild_prod = rebuild_prod[:, np.newaxis] * indus_shares  # type:ignore
        house_rebuild_prod = rebuild_prod[:, np.newaxis] * house_shares  # type:ignore

        if ind_reb:
            logger.debug(
                f"indus_rebuild_prod: {pd.DataFrame(indus_rebuild_prod, index=self.industries)}"
            )
        if h_reb:
            logger.debug(
                f"house_rebuild_prod: {pd.DataFrame(house_rebuild_prod, index=self.industries)}"
            )
            # logger.debug(f"tot_rebuilding_demand_summed: {pd.DataFrame(tot_rebuilding_demand_summed, index=self.industries)}")
        # logger.debug(f"house_rebuild_prod: {pd.Series(house_rebuild_prod.flatten(), index=self.industries)}")

        indus_rebuild_prod_distributed = np.zeros(shape=indus_reb_dem_per_event.shape)
        house_rebuild_prod_distributed = np.zeros(shape=house_reb_dem_per_event.shape)
        if ind_reb:
            # 1. We normalize rebuilding demand by total rebuilding demand (i.e. we get for each client asking, the share of the total demand)
            # This is done by broadcasting total demand and then dividing. (Perhaps this is not efficient ?)
            indus_rebuilding_demand_shares = np.zeros(
                shape=indus_reb_dem_per_event.shape
            )
            indus_rebuilding_demand_broad = np.broadcast_to(
                indus_reb_dem_tot_per_event[:, np.newaxis],
                indus_reb_dem_per_event.shape,
            )
            np.divide(
                indus_reb_dem_per_event,
                indus_rebuilding_demand_broad,
                where=(indus_rebuilding_demand_broad != 0),
                out=indus_rebuilding_demand_shares,
            )

            # 2. Then we multiply those shares by the total production (each client get production proportional to its demand relative to total demand)
            indus_rebuild_prod_broad = np.broadcast_to(
                indus_rebuild_prod[:, np.newaxis], indus_reb_dem_per_event.shape
            )
            np.multiply(
                indus_rebuilding_demand_shares,
                indus_rebuild_prod_broad,
                out=indus_rebuild_prod_distributed,
            )

        if h_reb:
            house_rebuilding_demand_shares = np.zeros(
                shape=house_reb_dem_per_event.shape
            )
            house_rebuilding_demand_broad = np.broadcast_to(
                house_reb_dem_tot_per_event[:, np.newaxis],
                house_reb_dem_per_event.shape,
            )
            np.divide(
                house_reb_dem_per_event,
                house_rebuilding_demand_broad,
                where=(house_rebuilding_demand_broad != 0),
                out=house_rebuilding_demand_shares,
            )
            house_rebuild_prod_broad = np.broadcast_to(
                house_rebuild_prod[:, np.newaxis], house_reb_dem_per_event.shape
            )
            np.multiply(
                house_rebuilding_demand_shares,
                house_rebuild_prod_broad,
                out=house_rebuild_prod_distributed,
            )

        # update rebuilding demand
        events_to_remove = []
        for e_id, evnt in enumerate(rebuildable_events):
            if len(evnt.rebuilding_demand_indus) > 0:
                evnt.rebuilding_demand_indus -= indus_rebuild_prod_distributed[
                    :, :, e_id
                ]
            if len(evnt.rebuilding_demand_house) > 0:
                evnt.rebuilding_demand_house -= house_rebuild_prod_distributed[
                    :, :, e_id
                ]
            if (evnt.rebuilding_demand_indus < (10 / self.monetary_factor)).all() and (
                evnt.rebuilding_demand_house < (10 / self.monetary_factor)
            ).all():
                events_to_remove.append(evnt)
        return events_to_remove

    def calc_matrix_stock_gap(self, matrix_stock_goal) -> npt.NDArray:
        """Computes and returns inputs stock gap matrix

        The gap is simply the difference between the goal (given as argument)
        and the current stock.

        Parameters
        ----------
        matrix_stock_goal : npt.NDArray of float
            The target inventories.

        Returns
        -------
        npt.NDArray
            The (only positive) gap between goal and current inventories.

        Raises
        ------
        RuntimeError
            If NaN are found in the result.

        """
        matrix_stock_gap = np.zeros(matrix_stock_goal.shape)
        # logger.debug("matrix_stock_goal: {}".format(matrix_stock_goal.shape))
        # logger.debug("matrix_stock: {}".format(self.matrix_stock.shape))
        # logger.debug("matrix_stock_goal_finite: {}".format(matrix_stock_goal[np.isfinite(matrix_stock_goal)].shape))
        # logger.debug("matrix_stock_finite: {}".format(self.matrix_stock[np.isfinite(self.matrix_stock)].shape))
        matrix_stock_gap[np.isfinite(matrix_stock_goal)] = (
            matrix_stock_goal[np.isfinite(matrix_stock_goal)]
            - self.matrix_stock[np.isfinite(self.matrix_stock)]
        )
        if np.isnan(matrix_stock_gap).any():
            raise RuntimeError("NaN in matrix stock gap")
        matrix_stock_gap[matrix_stock_gap < 0] = 0
        return matrix_stock_gap

    def calc_orders(self) -> None:
        r"""Computes and sets the orders (intermediate demand) for the next step.

        See :ref:`Order module <boario-math-orders>`

        Raises
        ------
        RuntimeError
            If negative orders are found, which shouldn't happen.
        """

        # total_demand = self.total_demand
        production_opt = self.production_opt.copy()
        matrix_stock_goal = np.tile(production_opt, (self.n_sectors, 1)) * self.tech_mat
        # Check this !
        with np.errstate(invalid="ignore"):
            matrix_stock_goal *= self.inv_duration[:, np.newaxis]
        if np.allclose(self.matrix_stock, matrix_stock_goal):
            matrix_stock_gap = np.zeros(matrix_stock_goal.shape)
        else:
            matrix_stock_gap = self.calc_matrix_stock_gap(matrix_stock_goal)
        matrix_stock_gap += (
            np.tile(self.production, (self.n_sectors, 1)) * self.tech_mat
        )
        if self.order_type == "alt":
            prod_ratio = np.ones(shape=self.X_0.shape)
            np.divide(
                self.production_cap, self.X_0, out=prod_ratio, where=self.X_0 != 0
            )
            Z_prod = self.Z_0 * prod_ratio[:, np.newaxis]
            Z_Cprod = np.tile(self._matrix_I_sum @ Z_prod, (self.n_regions, 1))
            out = np.zeros(shape=Z_prod.shape)
            np.divide(Z_prod, Z_Cprod, out=out, where=Z_Cprod != 0)
            tmp = np.tile(matrix_stock_gap, (self.n_regions, 1)) * out
        else:
            tmp = np.tile(matrix_stock_gap, (self.n_regions, 1)) * self.Z_distrib
        if (tmp < 0).any():
            raise RuntimeError("Negative orders computed")
        self.matrix_orders = tmp

    def calc_overproduction(self) -> None:
        r"""Computes and updates the overproduction vector.

        See :ref:`Overproduction module <boario-math-overprod>`

        """

        scarcity = np.full(self.production.shape, 0.0)
        total_demand = self.total_demand.copy()
        scarcity[total_demand != 0] = (
            total_demand[total_demand != 0] - self.production[total_demand != 0]
        ) / total_demand[total_demand != 0]
        scarcity[np.isnan(scarcity)] = 0
        overprod_chg = (
            ((self.overprod_max - self.overprod) * scarcity * self.overprod_tau)
            + (
                (self.overprod_base - self.overprod)
                * (scarcity == 0)
                * self.overprod_tau
            )
        ).flatten()
        self.overprod += overprod_chg
        self.overprod[self.overprod < 1.0] = 1.0

    def reset_module(
        self,
    ) -> None:
        """Resets the model to initial state [Deprecated]

        This method has not been checked extensively.
        """

        self.productive_capital_lost = np.zeros(self.production.shape)
        self.overprod = np.full(
            (self.n_regions * self.n_sectors), self.overprod_base, dtype=np.float64
        )
        self.matrix_stock = self.matrix_stock_0.copy()
        self.matrix_orders = self.Z_0.copy()
        self.production = self.X_0.copy()
        self.intmd_demand = self.Z_0.copy()
        self.final_demand = self.Y_0.copy()
        self.final_demand_not_met = np.zeros(self.Y_0.shape)
        self.rebuilding_demand = np.array([])
        self.in_shortage = False
        self.had_shortage = False
        self._prod_delta_type = None
        self._indus_rebuild_demand_tot = np.array([])
        self._house_rebuild_demand_tot = np.array([])
        self._indus_rebuild_demand = np.array([])
        self._house_rebuild_demand = np.array([])
        self._tot_rebuild_demand = np.array([])
        self._prod_cap_delta_productive_capital = np.array([])
        self._prod_cap_delta_arbitrary = np.array([])
        self._prod_cap_delta_tot = np.array([])

    def write_index(self, index_file: str | pathlib.Path) -> None:
        """Write the indexes of the different dataframes of the model in a json file.

        In order to easily rebuild the dataframes from the 'raw' data, this
        method create a JSON file with all columns and indexes names, namely :

        * regions names
        * sectors names
        * final demand categories
        * number of regions, sectors and industries (regions * sectors)

        Parameters
        ----------
        index_file : pathlib.Path
            Path to the file to save the indexes.
        """

        indexes = {
            "regions": list(self.regions),
            "sectors": list(self.sectors),
            "fd_cat": list(self.final_demand_cat),
            "n_sectors": self.n_sectors,
            "n_regions": self.n_regions,
            "n_industries": self.n_sectors * self.n_regions,
        }
        if isinstance(index_file, str):
            index_file = pathlib.Path(index_file)
        index_file.parent.mkdir(parents=True, exist_ok=True)
        with index_file.open("w") as ffile:
            json.dump(indexes, ffile)

    def change_inv_duration(self, new_dur, old_dur=None) -> None:
        # replace this method by a property !
        if old_dur is None:
            old_dur = self.main_inv_dur
        old_dur = float(old_dur) / self.n_temporal_units_by_step
        new_dur = float(new_dur) / self.n_temporal_units_by_step
        logger.info(
            "Changing (main) inventories duration from {} steps to {} steps (there are {} temporal units by step so duration is {})".format(
                old_dur,
                new_dur,
                self.n_temporal_units_by_step,
                new_dur * self.n_temporal_units_by_step,
            )
        )
        self.inv_duration = np.where(
            self.inv_duration == old_dur, new_dur, self.inv_duration
        )



# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Optional, Union

from boario.extended_models import ARIOPsiModel
from boario.simulation import Simulation
import numpyencoder
import json
import pathlib
import numpy as np
import pandas as pd
import itertools
from boario.utils import misc
import dask.dataframe as da
from boario import logger
from boario.event import *

__all__ = ["Indicators"]


def create_df(data, regions, sectors):
    df = pd.DataFrame(
        data,
        columns=pd.MultiIndex.from_product(
            [regions, sectors], names=["region", "sector"]
        ),
    )
    df = df.interpolate()
    df = df.rename_axis("step")
    return df


def df_from_memmap(memmap: np.memmap, indexes: dict) -> pd.DataFrame:
    a = pd.DataFrame(
        memmap,
        columns=pd.MultiIndex.from_product([indexes["regions"], indexes["sectors"]]),
    )
    a["step"] = a.index
    return a


def stock_df_from_memmap(
    memmap: np.memmap, indexes: dict, timesteps: int, timesteps_simulated: int
) -> pd.DataFrame:
    a = pd.DataFrame(
        memmap.reshape(timesteps * indexes["n_sectors"], -1),
        index=pd.MultiIndex.from_product(
            [timesteps, indexes["sectors"]], names=["step", "stock of"]
        ),
        columns=pd.MultiIndex.from_product([indexes["regions"], indexes["sectors"]]),
    )
    a = a.loc[pd.IndexSlice[:timesteps_simulated, :]]
    return a


class Indicators(object):
    record_files_list = [
        "final_demand",
        "intermediate_demand",
        "final_demand_unmet",
        "production_capacity",
        "production_realised",
        "limiting_inputs",
        "overproduction",
        "rebuild_demand",
        "rebuild_prod",
        "productive_capital_to_recover",
    ]

    params_list = ["simulated_params", "simulated_events"]

    def __init__(
        self,
        *,
        has_crashed,
        params,
        regions,
        sectors,
        productive_capital,
        prod,
        prodmax,
        overprod,
        final_demand,
        io_demand,
        r_demand,
        r_prod,
        fd_unmet,
        limiting_stocks,
        n_temporal_units_to_sim: int,
        n_temporal_units_by_step: int,
        events: list[Event],
        results_storage: Optional[str | pathlib.Path] = None,
        psi_param: Optional[float] = None,
        inventory_restoration_tau: Optional[int | list[int]] = None,
        stocks=None,
        stocks_treatment=False,
        include_crash: bool = False,
    ) -> None:
        logger.info("Instantiating indicators")
        if not include_crash:
            if has_crashed:
                raise RuntimeError(
                    "Simulation crashed and include_crash is False, I won't compute indicators"
                )
        steps = [i for i in range(n_temporal_units_to_sim)]

        self.params = params
        self.n_rows = n_temporal_units_to_sim
        self.n_temporal_units_by_step = n_temporal_units_by_step

        self.productive_capital_to_recover_df = create_df(
            productive_capital, regions, sectors
        )
        self.production_realised_df = create_df(prod, regions, sectors)
        self.production_capacity_df = create_df(prodmax, regions, sectors)
        self.overproduction_df = create_df(overprod, regions, sectors)
        self.final_demand_df = create_df(final_demand, regions, sectors)
        self.intermediate_demand_df = create_df(io_demand, regions, sectors)
        self.rebuild_demand_df = create_df(r_demand, regions, sectors)
        self.rebuild_prod_df = create_df(r_prod, regions, sectors)
        self.final_demand_unmet_df = create_df(fd_unmet, regions, sectors)

        if stocks_treatment:
            stocks_df = pd.DataFrame(
                stocks.reshape(n_temporal_units_to_sim * len(sectors), -1),
                index=pd.MultiIndex.from_product(
                    [steps, sectors], names=["step", "stock of"]
                ),
                columns=pd.MultiIndex.from_product(
                    [regions, sectors],
                    names=["region", "sector"],
                ),
            )
            stocks_df = stocks_df.replace([np.inf, -np.inf], np.nan).dropna(how="all")
            stocks_df = stocks_df.astype(np.float32)
            stocks_df = (
                stocks_df.groupby("stock of")
                .pct_change()
                .fillna(0)
                .add(1)
                .groupby("stock of")
                .cumprod()
                .sub(1)
            )  # type: ignore
            stocks_df = stocks_df.melt(ignore_index=False).rename(
                columns={
                    "variable_0": "region",
                    "variable_1": "sector",
                    "variable_2": "stock of",
                }
            )
            stocks_df = stocks_df.reset_index()
            stocks_df["step"] = stocks_df["step"].astype("uint16")
            stocks_df["stock of"] = stocks_df["stock of"].astype("category")
            stocks_df["region"] = stocks_df["region"].astype("category")
            stocks_df["sector"] = stocks_df["sector"].astype("category")

        else:
            stocks_df = None

        self.inputs_stocks_df = stocks_df
        del stocks_df
        # self.df_stocks = self.df_stocks.interpolate()

        self.treated_loss_df = (
            self.final_demand_unmet_df.melt(ignore_index=False)
            .rename(
                columns={
                    "variable_0": "region",
                    "variable_1": "fd_cat",
                    "value": "fdloss",
                }
            )
            .reset_index()
        )

        self.limiting_inputs_df = limiting_stocks
        # pd.DataFrame(
        #     limiting_stocks.reshape(n_temporal_units_to_sim * len(sectors), -1),
        #     index=pd.MultiIndex.from_product(
        #         [steps, sectors], names=["step", "stock of"]
        #     ),
        #     columns=pd.MultiIndex.from_product(
        #         [regions, sectors], names=["region", "sector"]
        #     ),
        # )
        self.aff_regions = []
        self.aff_sectors = []
        for e in events:
            if isinstance(e, dict):
                self.aff_sectors.append(e["aff_sectors"])
                self.aff_regions.append(e["aff_regions"])
            elif isinstance(e, Event):
                self.aff_sectors.append(e.aff_sectors)
                self.aff_regions.append(e.aff_regions)
            else:
                raise ValueError(f"Unrecognised event of type {type(e)}")

        self.aff_regions = list(misc.flatten(self.aff_regions))
        self.aff_sectors = list(misc.flatten(self.aff_sectors))

        self.rebuilding_sectors = []

        # This is working but wrong ! I need to fix it
        for e in events:
            if isinstance(e, dict):
                self.rebuilding_sectors.append(e.get("rebuilding_sectors"))
            if isinstance(e, EventKapitalRebuild):
                self.rebuilding_sectors.append(list(e.rebuilding_sectors))
            else:
                self.rebuilding_sectors = []

        self.rebuilding_sectors = list(misc.flatten(self.rebuilding_sectors))
        # This is also wrong but will settle for now
        gdp_dmg_share = -1.0
        self.indicators = {
            "region": self.aff_regions,
            "gdp_dmg_share": gdp_dmg_share,
            "tot_fd_unmet": "unset",
            "aff_fd_unmet": "unset",
            "rebuild_durations": "unset",
            "shortage_b": False,
            "shortage_date_start": "unset",
            "shortage_date_end": "unset",
            "shortage_date_max": "unset",
            "shortage_ind_max": "unset",
            "shortage_ind_mean": "unset",
            "10_first_shortages_(step,region,sector,stock_of)": "unset",
            "prod_gain_tot": "unset",
            "prod_lost_tot": "unset",
            "prod_gain_unaff": "unset",
            "prod_lost_unaff": "unset",
            "psi": psi_param,
            "inv_tau": inventory_restoration_tau,
            "n_temporal_units_to_sim": n_temporal_units_to_sim,
            "has_crashed": has_crashed,
        }
        if results_storage is not None:
            self.storage_path = (pathlib.Path(results_storage)).resolve() / "indicators"
            self.parquets_path = (pathlib.Path(results_storage)).resolve() / "parquets"
            self.storage = self.storage_path / "indicators.json"
            self.parquets_path.mkdir(parents=True, exist_ok=True)
            self.storage_path.mkdir(parents=True, exist_ok=True)
            self.save_dfs()

    @classmethod
    def from_simulation(
        cls,
        sim: Simulation,
        results_storage: Optional[str | pathlib.Path] = None,
        stocks_treatment: bool = False,
        include_crash: bool = False,
    ) -> Indicators:
        if isinstance(sim.model, ARIOPsiModel):
            psi = sim.model.psi
            inventory_restoration_tau = list(sim.model.restoration_tau)
        else:
            psi = None
            inventory_restoration_tau = None

        prod = getattr(sim, "production_realised")
        productive_capital = getattr(sim, "productive_capital_to_recover")
        prodmax = getattr(sim, "production_capacity")
        overprod = getattr(sim, "overproduction")
        final_demand = getattr(sim, "final_demand")
        io_demand = getattr(sim, "intermediate_demand")
        r_demand = getattr(sim, "rebuild_demand")
        r_prod = getattr(sim, "rebuild_prod")
        fd_unmet = getattr(sim, "final_demand_unmet")
        if stocks_treatment:
            stocks = getattr(sim, "inputs_stocks")
        else:
            stocks = None

        limiting_stocks = getattr(sim, "limiting_inputs")
        return cls(
            has_crashed=sim.has_crashed,
            params=sim.params_dict,
            regions=sim.model.regions,
            sectors=sim.model.sectors,
            productive_capital=productive_capital,
            prod=prod,
            prodmax=prodmax,
            overprod=overprod,
            final_demand=final_demand,
            io_demand=io_demand,
            r_demand=r_demand,
            r_prod=r_prod,
            fd_unmet=fd_unmet,
            limiting_stocks=limiting_stocks,
            n_temporal_units_to_sim=sim.n_temporal_units_to_sim,
            n_temporal_units_by_step=sim.model.n_temporal_units_by_step,
            events=sim.all_events,
            results_storage=results_storage,
            psi_param=psi,
            inventory_restoration_tau=inventory_restoration_tau,
            stocks=stocks,
            stocks_treatment=stocks_treatment,
            include_crash=include_crash,
        )

    @classmethod
    def from_folder(
        cls,
        folder: Union[str, pathlib.Path],
        include_crash: bool = False,
    ) -> Indicators:
        if not isinstance(folder, pathlib.Path):
            folder = pathlib.Path(folder)
            if not folder.exists():
                raise FileNotFoundError(str("Directory does not exist:" + str(folder)))

        with (folder / "jsons" / "indexes.json").open() as f:
            indexes = json.load(f)

        params_folder = folder / "jsons"
        records_folder = folder / "records"
        params_file = {f.stem: f for f in params_folder.glob("*.json")}
        absentee = [f for f in cls.params_list if f not in params_file.keys()]
        if absentee:
            raise FileNotFoundError(
                "Some of the required parameters files not found (looked for {} in {}".format(
                    cls.params_list, folder
                )
            )

        record_files = [f for f in records_folder.glob("*") if f.is_file()]
        absentee = [
            f
            for f in cls.record_files_list
            if f not in [fn.name for fn in record_files]
        ]
        if absentee:
            raise FileNotFoundError(
                "Some of the required records are not there : {}".format(absentee)
            )

        with params_file["simulated_params"].open("r") as f:
            params = json.load(f)

        with params_file["simulated_events"].open("r") as f:
            events = json.load(f)

        if "has_crashed" in params:
            has_crashed = params["has_crashed"]
        else:
            has_crashed = False

        results_storage = folder.absolute()
        records_path = records_folder.absolute()
        n_temporal_units_to_sim = params["n_temporal_units_to_sim"]
        t = n_temporal_units_to_sim
        n_temporal_units_by_step = params["n_temporal_units_by_step"]
        regions = indexes["regions"]
        sectors = indexes["sectors"]
        psi = params.get("psi_param")
        inventory_restoration_tau = params.get("inventory_restoration_tau")
        prod = np.memmap(
            records_path / "production_realised",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        productive_capital = np.memmap(
            records_path / "productive_capital_to_recover",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        prodmax = np.memmap(
            records_path / "production_capacity",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        overprod = np.memmap(
            records_path / "overproduction",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        final_demand = np.memmap(
            records_path / "final_demand",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        io_demand = np.memmap(
            records_path / "intermediate_demand",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        r_demand = np.memmap(
            records_path / "rebuild_demand",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        r_prod = np.memmap(
            records_path / "rebuild_prod",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_industries"]),
        )
        fd_unmet = np.memmap(
            records_path / "final_demand_unmet",
            mode="r+",
            dtype="float64",
            shape=(t, indexes["n_regions"] * indexes["n_sectors"]),
        )
        limiting_stocks = np.memmap(
            records_path / "limiting_inputs",
            mode="r+",
            dtype="byte",
            shape=(t * indexes["n_sectors"], indexes["n_industries"]),
        )
        if params.get("register_stocks", False):
            if not (records_path / "inputs_stocks").exists():
                raise FileNotFoundError(
                    "Stocks record file was not found {}".format(
                        records_path / "stocks_record"
                    )
                )
            stocks = np.memmap(
                records_path / "stocks_record",
                mode="r+",
                dtype="float64",
                shape=(t * indexes["n_sectors"], indexes["n_industries"]),
            )
            stocks_treatment = True
        else:
            stocks = None
            stocks_treatment = False

        return cls(
            has_crashed=has_crashed,
            params=params,
            regions=regions,
            sectors=sectors,
            productive_capital=productive_capital,
            prod=prod,
            prodmax=prodmax,
            overprod=overprod,
            final_demand=final_demand,
            io_demand=io_demand,
            r_demand=r_demand,
            r_prod=r_prod,
            fd_unmet=fd_unmet,
            limiting_stocks=limiting_stocks,
            n_temporal_units_to_sim=n_temporal_units_to_sim,
            n_temporal_units_by_step=n_temporal_units_by_step,
            events=events,
            results_storage=results_storage,
            psi_param=psi,
            inventory_restoration_tau=inventory_restoration_tau,
            stocks=stocks,
            stocks_treatment=stocks_treatment,
            include_crash=include_crash,
        )

    def calc_top_failing_sect(self):
        pass

    def calc_tot_fd_unmet(self):
        self.indicators["tot_fd_unmet"] = self.treated_loss_df["fdloss"].sum()

    def calc_aff_fd_unmet(self):
        # TODO: check this
        self.indicators["aff_fd_unmet"] = self.treated_loss_df[
            self.treated_loss_df.region.isin(self.aff_regions)
        ]["fdloss"].sum()

    def calc_rebuild_durations(self):
        rebuilding = (
            self.rebuild_demand_df.reset_index()
            .melt(
                id_vars="step",
                var_name=["region", "sector"],
                value_name="rebuild_demand",
            )
            .groupby("step")
            .sum(numeric_only=True)
            .ne(0)
            .rebuild_demand.to_numpy()
        )
        self.indicators["rebuild_durations"] = [
            sum(1 for _ in group) for key, group in itertools.groupby(rebuilding) if key
        ]

    def calc_recovery_duration(self):
        pass

    def calc_general_shortage(self):
        # TODO: replace hard values by soft.
        n_regions = self.limiting_inputs_df.columns.levels[0].size
        n_sectors = self.limiting_inputs_df.columns.levels[1].size
        # have only steps in index (next step not possible with multiindex)
        a = self.limiting_inputs_df.unstack()
        # select only simulated steps (we can't store nan in bool or byte dtype array)
        a = a.iloc[lambda x: x.index % self.n_temporal_units_by_step == 0]
        # We put -1 initially, so this should check we have correctly selected the simulated rows
        assert (a >= 0).all().all()
        # put input stocks as columns and the rest as index
        a = a.stack().T.stack(level=0)
        # sum for all input and divide by n_sector to get "input shortage fraction"
        # by industry
        b = a.sum(axis=1).groupby(["step", "region", "sector"]).sum() / n_sectors
        # by sector
        c = b.groupby(["step", "region"]).sum() / n_sectors
        # by region
        c = c.groupby("step").sum() / n_regions
        if not c.ne(0).any():
            self.indicators["shortage_b"] = False
        else:
            self.indicators["shortage_b"] = True
            shortage_date_start = c.ne(0.0).idxmax()
            self.indicators["shortage_date_start"] = shortage_date_start
            shortage_date_end = c.loc[shortage_date_start:].eq(0).idxmax()
            self.indicators["shortage_date_end"] = shortage_date_end
            self.indicators["shortage_date_max"] = c.idxmax()
            self.indicators["shortage_ind_max"] = c.max()
            self.indicators["shortage_ind_mean"] = c.loc[
                shortage_date_start:shortage_date_end
            ].mean()

    def calc_first_shortages(self):
        a = self.limiting_inputs_df.stack([0, 1])  # type: ignore
        a = a.swaplevel(1, 2).swaplevel(2, 3)
        b = a[a > 0]
        b = b[:10]
        res = list(b.index)  # type:ignore
        self.indicators["10_first_shortages_(step,region,sector,stock_of)"] = res

    def calc_tot_prod_change(self):
        df2 = self.production_realised_df.copy()
        # df2.columns=df2.columns.droplevel(0)
        prod_chg = df2 - df2.iloc[0, :]
        # Round to € to avoid floating error differences
        prod_chg = prod_chg.round(6)
        # Aggregate rebuilding and non-rebuilding sectors
        prod_chg_agg_1 = pd.concat(
            [
                prod_chg.loc[:, pd.IndexSlice[:, self.rebuilding_sectors]]
                .groupby("region", axis=1)
                .sum(),
                prod_chg.loc[
                    prod_chg.index.difference(
                        prod_chg.loc[
                            :, pd.IndexSlice[:, self.rebuilding_sectors]
                        ].columns
                    )
                ]
                .groupby("region", axis=1)
                .sum(),
            ],
            keys=["rebuilding", "non-rebuilding"],
            names=["sectors affected"],
            axis=1,
        )

        n_semesters = self.n_rows // (self.params["year_to_temporal_unit_factor"] // 2)
        row_to_semester = self.params["year_to_temporal_unit_factor"] // 2
        modulo = self.params["year_to_temporal_unit_factor"] % 2
        logger.info(
            "There are {} semesters [{} rows, each representing a {}th of a year]".format(
                n_semesters, self.n_rows, self.params["year_to_temporal_unit_factor"]
            )
        )
        prod_chg_sect_sem_l = []
        for sem in range(0, n_semesters):
            prod_chg_sect_sem_l.append(
                prod_chg_agg_1.iloc[
                    sem * row_to_semester : ((sem + 1) * row_to_semester)
                    + (sem % 2) * modulo
                ]
                .sum()
                .T
            )

        prod_chg_region = pd.concat(
            prod_chg_sect_sem_l,
            keys=["semester {}".format(v + 1) for v in range(n_semesters)],
            names=["semester"],
        )
        aff_regions = "~".join(self.aff_regions)
        self.prod_chg_region = pd.DataFrame({aff_regions: prod_chg_region}).T
        prod_chg_sect = prod_chg.sum()
        tmp = prod_chg_sect.sort_values(ascending=False, key=abs).head(5).to_dict()
        self.indicators["top_5_sector_chg"] = {str(k): v for k, v in tmp.items()}

        self.indicators["prod_gain_tot"] = prod_chg.mul(prod_chg.gt(0)).sum().sum()
        self.indicators["prod_lost_tot"] = prod_chg.mul(~prod_chg.gt(0)).sum().sum() * (
            -1
        )
        prod_chg = prod_chg.drop(self.aff_regions, axis=1)
        self.indicators["prod_gain_unaff"] = prod_chg.mul(prod_chg.gt(0)).sum().sum()
        self.indicators["prod_lost_unaff"] = prod_chg.mul(
            ~prod_chg.gt(0)
        ).sum().sum() * (-1)

    def update_indicators(self):
        logger.info("(Re)computing all indicators")
        logger.info("Tot fd unmet")
        self.calc_tot_fd_unmet()
        logger.info("aff fd unmet")
        self.calc_aff_fd_unmet()
        logger.info("rebuild durations")
        self.calc_rebuild_durations()
        logger.info("recovery duration")
        self.calc_recovery_duration()
        logger.info("general shortage")
        self.calc_general_shortage()
        logger.info("tot prod change")
        self.calc_tot_prod_change()
        logger.info("fd loss region")
        self.calc_fd_loss_region()
        logger.info("first shortages")
        self.calc_first_shortages()

    def write_indicators(self, storage_path=None):
        if storage_path is None and not hasattr(self, "storage_path"):
            raise ValueError(
                f"You are attempting to save indicators but no storage path was specified either in the Indicators class or to this method."
            )
        storage_path = self.storage_path if storage_path is None else storage_path
        storage_path = pathlib.Path(storage_path).resolve()
        logger.info("Writing indicators to json")
        if hasattr(self, "prod_chg_region"):
            self.prod_chg_region.to_json(
                storage_path / "prod_chg.json", indent=4, orient="split"
            )
        if hasattr(self, "fd_loss_region"):
            self.fd_loss_region.to_json(
                storage_path / "fd_loss.json", indent=4, orient="split"
            )
        with (storage_path / "indicators.json").open("w") as f:
            json.dump(self.indicators, f, cls=numpyencoder.NumpyEncoder, indent=4)

    def calc_fd_loss_region(self):
        fd_loss = self.final_demand_unmet_df.copy().round(6)
        fd_loss_agg_1 = pd.concat(
            [
                fd_loss.loc[:, pd.IndexSlice[:, self.rebuilding_sectors]]
                .groupby("region", axis=1)
                .sum(),
                fd_loss.loc[
                    fd_loss.index.difference(
                        fd_loss.loc[
                            :, pd.IndexSlice[:, self.rebuilding_sectors]
                        ].columns
                    )
                ]
                .groupby("region", axis=1)
                .sum(),
            ],
            keys=["rebuilding", "non-rebuilding"],
            names=["sectors affected"],
            axis=1,
        )

        n_semesters = self.n_rows // (self.params["year_to_temporal_unit_factor"] // 2)
        row_to_semester = self.params["year_to_temporal_unit_factor"] // 2
        modulo = self.params["year_to_temporal_unit_factor"] % 2
        logger.info(
            "There are {} semesters [{} rows, each representing a {}th of a year]".format(
                n_semesters, self.n_rows, self.params["year_to_temporal_unit_factor"]
            )
        )
        fd_loss_sect_sem_l = []
        for sem in range(0, n_semesters):
            fd_loss_sect_sem_l.append(
                fd_loss_agg_1.iloc[
                    sem * row_to_semester : ((sem + 1) * row_to_semester)
                    + (sem % 2) * modulo
                ]
                .sum()
                .T
            )

        fd_loss_region = pd.concat(
            fd_loss_sect_sem_l,
            keys=["semester {}".format(v + 1) for v in range(n_semesters)],
            names=["semester"],
        )
        aff_regions = "~".join(self.aff_regions)
        self.fd_loss_region = pd.DataFrame({aff_regions: fd_loss_region}).T
        fd_loss_sect = fd_loss.sum()
        tmp = fd_loss_sect.sort_values(ascending=False, key=abs).head(5).to_dict()
        self.indicators["top_5_sector_fdloss"] = {str(k): v for k, v in tmp.items()}

    def save_dfs(self):
        logger.info("Saving computed dataframe to results folder")
        self.production_realised_df.to_parquet(
            self.parquets_path / "production_realised_df.parquet"
        )
        self.productive_capital_to_recover_df.to_parquet(
            self.parquets_path / "productive_capital_to_recover_df.parquet"
        )
        self.production_capacity_df.to_parquet(
            self.parquets_path / "production_capacity_df.parquet"
        )
        self.overproduction_df.to_parquet(
            self.parquets_path / "overproduction_df.parquet"
        )
        self.final_demand_unmet_df.to_parquet(
            self.parquets_path / "final_demand_unmet_df.parquet"
        )
        self.final_demand_df.to_parquet(self.parquets_path / "final_demand_df.parquet")
        self.intermediate_demand_df.to_parquet(
            self.parquets_path / "intermediate_demand_df.parquet"
        )
        self.rebuild_demand_df.to_parquet(
            self.parquets_path / "rebuild_demand_df.parquet"
        )
        self.treated_loss_df.to_parquet(self.parquets_path / "treated_df_loss.parquet")
        if self.inputs_stocks_df is not None:
            ddf = da.from_pandas(self.inputs_stocks_df, chunksize=10000000)
            ddf.to_parquet(
                self.parquets_path / "treated_df_stocks.parquet", engine="pyarrow"
            )
        if self.limiting_inputs_df is not None:
            # ddf_l = da.from_pandas(self.df_limiting, chunksize=10000000)
            # ddf_l = ddf_l.melt(ignore_index=False).rename(columns={'variable_0':'region','variable_1':'sector', 'variable_2':'stock of'})
            # ddf_l = ddf_l.reset_index()
            # ddf_l['step'] = ddf_l['step'].astype("uint16")
            # ddf_l['stock of'] = ddf_l['stock of'].astype("category")
            # ddf_l['region'] = ddf_l['region'].astype("category")
            # ddf_l['sector'] = ddf_l['sector'].astype("category")
            self.limiting_inputs_df.to_parquet(
                self.parquets_path / "treated_df_limiting_inputs.parquet",
                engine="pyarrow",
            )
        # self.df_limiting.to_feather(self.parquets_path/"treated_df_limiting.feather")
